# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from datetime import date
from pathlib import Path
from typing import Any, Dict, List, Optional, Set

import ezodf

from rp2.abstract_country import AbstractCountry
from rp2.abstract_report_generator import AbstractReportGenerator
from rp2.abstract_transaction import AbstractTransaction
from rp2.computed_data import ComputedData
from rp2.configuration import MAX_DATE, MIN_DATE, Configuration
from rp2.in_transaction import InTransaction
from rp2.localization import _
from rp2.out_transaction import OutTransaction
from rp2.rp2_decimal import RP2Decimal
from rp2.rp2_error import RP2RuntimeError, RP2TypeError


class AbstractODSGenerator(AbstractReportGenerator):
    @classmethod
    def _initialize_output_file(  # pylint: disable=too-many-branches
        cls,
        country: AbstractCountry,  # pylint: disable=unused-argument
        legend_data: List[List[str]],
        years_2_accounting_method_names: Dict[int, str],
        output_dir_path: str,
        output_file_prefix: str,
        output_file_name: str,
        template_path: str,
        template_sheets_to_keep: Set[str],
        from_date: date,
        to_date: date,
    ) -> Any:
        Configuration.type_check_string("output_dir_path", output_dir_path)
        Configuration.type_check_string("output_file_prefix", output_file_prefix)
        Configuration.type_check_string("output_file_name", output_file_name)
        if not isinstance(template_sheets_to_keep, Set):
            raise RP2TypeError(f"Parameter 'template_sheets_to_keep' is not a Set: {template_sheets_to_keep}")

        accounting_method: str = years_2_accounting_method_names[MIN_DATE.year] if len(years_2_accounting_method_names) == 1 else "mixed"
        output_file_path: Path = Path(output_dir_path) / Path(f"{output_file_prefix}{accounting_method}_{output_file_name}")
        if Path(output_file_path).exists():
            output_file_path.unlink()

        output_file: Any = ezodf.newdoc("ods", str(output_file_path), template=template_path)
        legend_sheet_name: str = f"__Legend_{cls.get_name()}"
        template_sheets_to_keep_with_legend: Set[str] = template_sheets_to_keep.copy()
        template_sheets_to_keep_with_legend.add(legend_sheet_name)

        index: int = 0
        sheet_name: str
        sheet_indexes_to_remove: List[int] = []
        for sheet_name in output_file.sheets.names():
            if sheet_name in template_sheets_to_keep_with_legend:
                if not sheet_name.startswith("__"):
                    raise RP2RuntimeError(f"Internal error: template sheet '{sheet_name}' doesn't start with '__'")
                # Template sheets' names start with "__": remove leading "__" from name of sheet we want to keep
                output_file.sheets[index].name = sheet_name[2:]
            elif sheet_name.startswith("__"):
                # Template sheet we don't want to keep: mark it for removal
                sheet_indexes_to_remove.append(index)
            index += 1

        # Setup legend sheet
        legend_sheet: Any = output_file.sheets[legend_sheet_name[2:]]
        cls._fill_page(legend_data, legend_sheet, 0, 0)
        index = 0
        method_cell_found: bool = False
        for index in range(0, 100):
            if legend_sheet[index, 0].value == _("Accounting Method"):
                accounting_method_by_year: List[str] = []
                if len(years_2_accounting_method_names) == 1:
                    accounting_method_by_year.append(years_2_accounting_method_names[MIN_DATE.year].upper())
                else:
                    old_year = MIN_DATE.year
                    for year, method in years_2_accounting_method_names.items():
                        if year - old_year > 1:
                            accounting_method_by_year.append(f"{old_year}->{year}:{method.upper()}")
                        else:
                            accounting_method_by_year.append(f"{year}:{method.upper()}")
                        old_year = year
                cls._fill_cell(legend_sheet, index, 1, ", ".join(accounting_method_by_year), visual_style="transparent")
                method_cell_found = True
                cls._fill_cell(legend_sheet, index + 1, 1, from_date if from_date != MIN_DATE else "non-specified", visual_style="transparent")
                cls._fill_cell(legend_sheet, index + 2, 1, to_date if to_date != MAX_DATE else "non-specified", visual_style="transparent")
                break
        if not method_cell_found:
            raise RP2RuntimeError(f"Internal error: ODS template has no '{_('Accounting Method')}' cell in column 0 of '{_('Legend')}'' sheet")

        legend_sheet.name = _("Legend")

        # Remove sheets that were marked for removal
        for index in reversed(sheet_indexes_to_remove):
            del output_file.sheets[index]

        return output_file

    def _get_template_path(self, template_name: str, country: Optional[AbstractCountry], generation_language: str) -> str:
        country_path = f"{country.country_iso_code}/" if country else ""
        language_suffix = f"_{generation_language}" if country else ""
        base_path = Path(os.path.dirname(__file__)).absolute() / Path(f"data/{country_path}template_{template_name}{language_suffix}")
        ods_path = Path(f"{base_path}.ods")
        if ods_path.exists():
            return str(ods_path)
        if not country:
            raise RP2RuntimeError(f"Internal error: template {ods_path} doesn't exist")

        # Look for a link (a .txt file containing the path to the .ods file)
        txt_path = Path(f"{base_path}.txt")
        if txt_path.exists():
            new_ods_path = Path(os.path.dirname(__file__)).absolute()
            with open(txt_path, encoding="utf-8") as template_link:
                contents = template_link.read().strip()
                if not contents or not contents.endswith(".ods"):
                    raise RP2RuntimeError(f"Internal error: template link {txt_path} doesn't contain a path ending with .ods: {contents}")
                new_ods_path = new_ods_path / Path(f"data/{contents}")
                if new_ods_path.exists():
                    return str(new_ods_path)
            raise RP2RuntimeError(f"Internal error: template link {txt_path} points to a path that doesn't exist: {new_ods_path}")

        raise RP2RuntimeError(f"Language {generation_language} not supported for country {country.country_iso_code}: template {ods_path}/.txt doesn't exist")

    def generate(
        self,
        country: AbstractCountry,
        years_2_accounting_method_names: Dict[int, str],
        asset_to_computed_data: Dict[str, ComputedData],
        output_dir_path: str,
        output_file_prefix: str,
        from_date: date,
        to_date: date,
        generation_language: str,
    ) -> None:
        raise NotImplementedError("Abstract method: it must be implemented in the plugin class")

    @staticmethod
    def _apply_style_to_cell(sheet: Any, row_index: int, column_index: int, style_name: str) -> None:
        Configuration.type_check_positive_int("row_index", row_index)
        Configuration.type_check_positive_int("column_index", column_index)
        Configuration.type_check_string("style_name", style_name)

        sheet[row_index, column_index].style_name = style_name

    @classmethod
    def _fill_cell(
        cls,
        sheet: Any,
        row_index: int,
        column_index: int,
        value: Any,
        visual_style: str = "transparent",
        data_style: str = "default",
        apply_style: bool = True,
    ) -> None:
        Configuration.type_check_string("visual_style", visual_style)
        Configuration.type_check_string("data_style", data_style)

        is_formula: bool = False
        if isinstance(value, str) and value and value[0] == "=":
            # If the value starts with '=' it is assumed to be a formula
            is_formula = True

        style_name: str = f"{visual_style}_{data_style}"
        if isinstance(value, RP2Decimal):
            # The ezodf API doesn't accept RP2Decimal, so we are forced to cast to float before writing to the spreadsheet
            value = float(value)
        if is_formula:
            sheet[row_index, column_index].formula = value
        else:
            sheet[row_index, column_index].set_value(value)
        if apply_style:
            cls._apply_style_to_cell(sheet=sheet, row_index=row_index, column_index=column_index, style_name=style_name)

    def _fill_header(
        self, title: str, header_row_1: List[str], header_row_2: List[str], sheet: Any, row_index: int, column_index: int, apply_style: bool = True
    ) -> int:
        Configuration.type_check_string("title", title)
        if not isinstance(header_row_1, List):
            raise RP2TypeError("Parameter 'header_row_1' is not a List")
        if not isinstance(header_row_2, List):
            raise RP2TypeError("Parameter 'header_row_2' is not a List")

        self._fill_cell(sheet, row_index, 0, title, visual_style="title", apply_style=apply_style)
        row_index += 1

        self._fill_cell(sheet, row_index, 0, "", visual_style="transparent", apply_style=apply_style)
        self._fill_cell(sheet, row_index + 1, 0, "", visual_style="transparent", apply_style=apply_style)

        header1: str
        header2: str
        i: int = 0
        for header1, header2 in zip(header_row_1, header_row_2):
            self._fill_cell(sheet, row_index, column_index + i, header1, visual_style="header", data_style="default", apply_style=apply_style)
            self._fill_cell(sheet, row_index + 1, column_index + i, header2, visual_style="header", data_style="default", apply_style=apply_style)
            i += 1
        return row_index + 2

    @classmethod
    def _fill_page(cls, data: List[List[str]], sheet: Any, row_index: int, column_index: int) -> int:
        if not isinstance(data, List):
            raise RP2TypeError("Parameter 'data' is not a List")

        i: int = 0
        for row in data:
            j: int = 0
            for element in row:
                # Apply no style because we don't know the nature of the data
                cls._fill_cell(sheet, row_index + i, column_index + j, element, apply_style=False)
                j += 1
            i += 1

        return row_index + i

    @staticmethod
    def _get_table_type_from_transaction(transaction: AbstractTransaction) -> str:
        AbstractTransaction.type_check("transaction", transaction)
        if isinstance(transaction, InTransaction):
            return "IN"
        if isinstance(transaction, OutTransaction):
            return "OUT"
        return "INTRA"



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

CONFIGURATION_SCHEMA = {
    "type": "object",
    "properties": {
        "in_header": {
            "type": "object",
            "properties": {
                "timestamp": {
                    "type": "integer",
                    "minimum": 0,
                },
                "asset": {
                    "type": "integer",
                    "minimum": 0,
                },
                "exchange": {
                    "type": "integer",
                    "minimum": 0,
                },
                "holder": {
                    "type": "integer",
                    "minimum": 0,
                },
                "transaction_type": {
                    "type": "integer",
                    "minimum": 0,
                },
                "spot_price": {
                    "type": "integer",
                    "minimum": 0,
                },
                "crypto_in": {
                    "type": "integer",
                    "minimum": 0,
                },
                "crypto_fee": {
                    "type": "integer",
                    "minimum": 0,
                },
                "fiat_in_no_fee": {
                    "type": "integer",
                    "minimum": 0,
                },
                "fiat_in_with_fee": {
                    "type": "integer",
                    "minimum": 0,
                },
                "fiat_fee": {
                    "type": "integer",
                    "minimum": 0,
                },
                "unique_id": {
                    "type": "integer",
                    "minimum": 0,
                },
                "notes": {
                    "type": "integer",
                    "minimum": 0,
                },
            },
            "required": [
                "timestamp",
                "asset",
                "exchange",
                "holder",
                "transaction_type",
                "spot_price",
                "crypto_in",
            ],
            "additionalProperties": False,
        },
        "out_header": {
            "type": "object",
            "properties": {
                "timestamp": {
                    "type": "integer",
                    "minimum": 0,
                },
                "asset": {
                    "type": "integer",
                    "minimum": 0,
                },
                "exchange": {
                    "type": "integer",
                    "minimum": 0,
                },
                "holder": {
                    "type": "integer",
                    "minimum": 0,
                },
                "transaction_type": {
                    "type": "integer",
                    "minimum": 0,
                },
                "spot_price": {
                    "type": "integer",
                    "minimum": 0,
                },
                "crypto_out_no_fee": {
                    "type": "integer",
                    "minimum": 0,
                },
                "crypto_fee": {
                    "type": "integer",
                    "minimum": 0,
                },
                "crypto_out_with_fee": {
                    "type": "integer",
                    "minimum": 0,
                },
                "fiat_out_no_fee": {
                    "type": "integer",
                    "minimum": 0,
                },
                "fiat_fee": {
                    "type": "integer",
                    "minimum": 0,
                },
                "unique_id": {
                    "type": "integer",
                    "minimum": 0,
                },
                "notes": {
                    "type": "integer",
                    "minimum": 0,
                },
            },
            "required": [
                "timestamp",
                "asset",
                "exchange",
                "holder",
                "transaction_type",
                "spot_price",
                "crypto_out_no_fee",
                "crypto_fee",
            ],
            "additionalProperties": False,
        },
        "intra_header": {
            "type": "object",
            "properties": {
                "timestamp": {
                    "type": "integer",
                    "minimum": 0,
                },
                "asset": {
                    "type": "integer",
                    "minimum": 0,
                },
                "from_exchange": {
                    "type": "integer",
                    "minimum": 0,
                },
                "from_holder": {
                    "type": "integer",
                    "minimum": 0,
                },
                "to_exchange": {
                    "type": "integer",
                    "minimum": 0,
                },
                "to_holder": {
                    "type": "integer",
                    "minimum": 0,
                },
                "spot_price": {
                    "type": "integer",
                    "minimum": 0,
                },
                "crypto_sent": {
                    "type": "integer",
                    "minimum": 0,
                },
                "crypto_received": {
                    "type": "integer",
                    "minimum": 0,
                },
                "unique_id": {
                    "type": "integer",
                    "minimum": 0,
                },
                "notes": {
                    "type": "integer",
                    "minimum": 0,
                },
            },
            "required": [
                "timestamp",
                "asset",
                "from_exchange",
                "from_holder",
                "to_exchange",
                "to_holder",
                "spot_price",
                "crypto_sent",
                "crypto_received",
            ],
            "additionalProperties": False,
        },
        "assets": {
            "type": "array",
            "items": {
                "type": "string",
                "minLength": 1,
            },
            "minItems": 1,
            "uniqueItems": True,
        },
        "exchanges": {
            "type": "array",
            "items": {
                "type": "string",
                "minLength": 1,
            },
            "minItems": 1,
            "uniqueItems": True,
        },
        "holders": {
            "type": "array",
            "items": {
                "type": "string",
                "minLength": 1,
            },
            "minItems": 1,
            "uniqueItems": True,
        },
        "generators": {
            "type": "array",
            "items": {
                "type": "string",
                "minLength": 1,
            },
            "minItems": 0,
            "uniqueItems": True,
        },
        "accounting_methods": {
            "type": "object",
            "propertyNames": {"pattern": "[2-9][0-9][0-9][0-9]"},
            "patternProperties": {"[2-9][0-9][0-9][0-9]": {"type": "string"}},
            "additionalProperties": False,
        },
    },
}



# Copyright 2022 mdavid217
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from datetime import date
from pathlib import Path
from typing import Any, Dict, List, Set, cast

from rp2.abstract_country import AbstractCountry
from rp2.computed_data import ComputedData
from rp2.in_transaction import InTransaction
from rp2.localization import _
from rp2.logger import create_logger
from rp2.plugin.report.abstract_ods_generator import AbstractODSGenerator
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2TypeError

LOGGER: logging.Logger = create_logger("open_positions")

_TEMPLATE_SHEETS: Set[str] = {"Asset", "Asset - Exchange", "Input"}
_TEMPLATE_SHEETS_TO_KEEP: Set[str] = {"__" + sheet_name for sheet_name in _TEMPLATE_SHEETS}
_FIAT_UNIT_DATA_STYLE_2_DECIMAL_MINIMUM = RP2Decimal("1")
_FIAT_UNIT_DATA_STYLE_4_DECIMAL_MINIMUM = RP2Decimal("0.20")

_ASSET: str = "Asset"
_ASSET_EXCHANGE: str = "Asset - Exchange"
_INPUT: str = "Input"
_INPUT_VALUE_STRING: str = "Enter asset value"
_REPORT_INPUT_VALUE_STRING: str = "See Input tab"


class Generator(AbstractODSGenerator):
    OUTPUT_FILE: str = "open_positions.ods"
    HEADER_ROWS = 3

    __legend: List[List[str]] = []
    __asset_header_names_row_1: List[str] = []
    __asset_header_names_row_2: List[str] = []
    __asset_exchange_header_names_row_1: List[str] = []
    __asset_exchange_header_names_row_2: List[str] = []
    __input_header_names_row_1: List[str] = []
    __input_header_names_row_2: List[str] = []

    # pylint: disable=line-too-long
    def _setup_text_data(self, country: AbstractCountry) -> None:
        currency_code: str = country.currency_iso_code.upper()

        self.__legend: List[List[str]] = [
            # fmt: off
            [_("Open Positions / Unrealized Gains")],
            [_("Fill in Asset Prices in the Input Tab for calculations")],
            [_("This report leverages the details of your transactions to give you a snapshot of the value of your unsold holdings")],
            [_("This file contains two versions of the same report. One version is grouped by Asset, the other by Asset and Exchange (or wallet)")],
            [""],
            [_("General")],
            [_("Accounting Method")],
            [_("From Date Filter")],
            [_("To Date Filter")],
            [""],
            [_("Report Fields")],
            [_("Asset"), _("Crypto / asset symbol")],
            [_("Holder"), _("Asset holder")],
            [_("Exchange"), _("Exchange or wallet where balance is held (Asset - Exchange tab only)")],
            [_("Crypto Balance"), _("Amount of given asset held")],
            [_("{} Per Unit Cost Basis").format(currency_code), _("Sum of fiat cost to acquire the balance of crypto based on in-flow transactions divided by the crypto balance")],
            [_("{} Unrealized Cost Basis").format(currency_code), _("Sum of fiat cost to acquire the balance of crypto based on in-flow transactions")],
            [_("Cost Basis Weight %"), _("Size of investment relative to other assets for cost basis ({} Unrealized Cost Basis divided by the sum of all assets' {} Unrealized Cost Basis)").format(currency_code, currency_code)],
            [_("{} Per Unit Input Price").format(currency_code), _("Looked up from the Input tab for the given asset")],
            [_("{} Unrealized Value").format(currency_code), _("Crypto Balance * Input Price")],
            [_("{} Unrealized Gain / Loss").format(currency_code), _("Unrealized Value - Unrealized Cost Basis")],
            [_("{} Unrealized Gain / Loss %").format(currency_code), _("Percent change between Unrealized Value and Unrealized Cost Basis")],
            [_("Unrealized G/L of Total Cost Basis %"), _("Indicates the percent of which this asset's unrealized gain/loss contributed to the whole portfolio's Unrealized Gain / Loss % (Gain/Loss relative to total cash investment)")],
            [_("{} Unrealized Value Weight %").format(currency_code), _("Size of investment relative to the portfolio's unrealized value (Unrealized Value divided by sum of all assets' Unrealized Values)")],
            [""],
            [_("Color Code")],
            [_("Gray"), _("Information from transactions")],
            [_("Yellow"), _("Information calculated based on values entered in the Input tab")],
            # fmt: on
        ]

        self.__asset_header_names_row_1: List[str] = [
            "",
            "",
            _("Crypto"),
            _("{} Per Unit").format(currency_code),
            _("{} Unrealized").format(currency_code),
            _("Cost Basis"),
            _("{} Per Unit").format(currency_code),
            _("{} Unrealized").format(currency_code),
            _("{} Unrealized").format(currency_code),
            _("{} Unrealized").format(currency_code),
            _("Unrealized G/L of"),
            _("{} Unrealized").format(currency_code),
        ]

        self.__asset_header_names_row_2: List[str] = [
            _("Asset"),
            _("Holder"),
            _("Balance"),
            _("Cost Basis"),
            _("Cost Basis"),
            _("Weight %"),
            _("Input Price"),
            _("Value"),
            _("Gain / Loss"),
            _("Gain / Loss %"),
            _("Total Cost Basis %"),
            _("Value Weight %"),
        ]

        self.__asset_exchange_header_names_row_1: List[str] = [
            "",
            "",
            "",
            _("Crypto"),
            _("{} Per Unit").format(currency_code),
            _("{} Unrealized").format(currency_code),
            _("Cost Basis"),
            _("{} Per Unit").format(currency_code),
            _("{} Unrealized").format(currency_code),
            _("{} Unrealized").format(currency_code),
            _("{} Unrealized").format(currency_code),
            _("Unrealized G/L of"),
            _("{} Unrealized").format(currency_code),
        ]

        self.__asset_exchange_header_names_row_2: List[str] = [
            _("Asset"),
            _("Holder"),
            _("Exchange"),
            _("Balance"),
            _("Cost Basis"),
            _("Cost Basis"),
            _("Weight %"),
            _("Input Price"),
            _("Value"),
            _("Gain / Loss"),
            _("Gain / Loss %"),
            _("Total Cost Basis %"),
            _("Value Weight %"),
        ]

        self.__input_header_names_row_1: List[str] = [
            _("Crypto"),
            "",
        ]

        self.__input_header_names_row_2: List[str] = [
            _("Asset"),
            _("Price"),
        ]

    def generate(
        self,
        country: AbstractCountry,
        years_2_accounting_method_names: Dict[int, str],
        asset_to_computed_data: Dict[str, ComputedData],
        output_dir_path: str,
        output_file_prefix: str,
        from_date: date,
        to_date: date,
        generation_language: str,
    ) -> None:
        # pylint: disable=too-many-branches

        if not isinstance(asset_to_computed_data, Dict):
            raise RP2TypeError(f"Parameter 'asset_to_computed_data' has non-Dict value {asset_to_computed_data}")

        self._setup_text_data(country)

        template_path: str = self._get_template_path("open_positions", country, generation_language)

        output_file: Any
        output_file = self._initialize_output_file(
            country=country,
            legend_data=self.__legend,
            years_2_accounting_method_names=years_2_accounting_method_names,
            output_dir_path=output_dir_path,
            output_file_prefix=output_file_prefix,
            output_file_name=self.OUTPUT_FILE,
            template_path=template_path,
            template_sheets_to_keep=_TEMPLATE_SHEETS_TO_KEEP,
            from_date=from_date,
            to_date=to_date,
        )

        asset: str
        computed_data: ComputedData

        asset_sheet = output_file.sheets[_ASSET]
        asset_exchange_sheet = output_file.sheets[_ASSET_EXCHANGE]
        input_sheet = output_file.sheets[_INPUT]

        self._fill_header(_("Open Positions by Asset"), self.__asset_header_names_row_1, self.__asset_header_names_row_2, asset_sheet, 0, 0, apply_style=False)
        self._fill_cell(asset_sheet, 0, 6, _("ENTER PRICES ON INPUT TAB"), apply_style=False)

        self._fill_header(
            _("Open Positions by Asset and Exchange"),
            self.__asset_exchange_header_names_row_1,
            self.__asset_exchange_header_names_row_2,
            asset_exchange_sheet,
            0,
            0,
            apply_style=False,
        )
        self._fill_cell(asset_exchange_sheet, 0, 7, _("ENTER PRICES ON INPUT TAB"), apply_style=False)

        self._fill_header(_("Asset Price Lookup Table"), self.__input_header_names_row_1, self.__input_header_names_row_2, input_sheet, 0, 0, apply_style=False)

        row_indexes: Dict[str, int] = {sheet_name: self.HEADER_ROWS for sheet_name in _TEMPLATE_SHEETS}

        # First loop - primary data collection
        #  - total_cost_basis: total cost basis for all transactions and all assets
        #  - asset_cost_bases: total cost basis for all transactions by asset
        #  - holders: list of holders with non-zero balances.
        #  - asset_crypto_balance_holder: sum of crypto balance for each asset by holder
        #  - asset_crypto_balance_holder_exchange: sum of crypto balance for each asset by holder and exchange
        total_cost_basis = ZERO
        asset_cost_bases: Dict[str, RP2Decimal] = {}
        holders: List[str] = []
        asset_crypto_balance_holder: Dict[str, Dict[str, RP2Decimal]] = {}
        asset_crypto_balance_holder_exchange: Dict[str, Dict[str, Dict[str, RP2Decimal]]] = {}

        for asset, computed_data in asset_to_computed_data.items():
            if not isinstance(asset, str):
                raise RP2TypeError(f"Parameter 'asset' has non-string value {asset}")
            ComputedData.type_check("computed_data", computed_data)

            # process in-flow transactions to collect the fiat cost basis data.
            for current_transaction in computed_data.in_transaction_set:
                in_transaction = cast(InTransaction, current_transaction)
                sold_percent: RP2Decimal = computed_data.get_in_lot_sold_percentage(in_transaction)
                transaction_cost_basis: RP2Decimal = in_transaction.fiat_in_with_fee * (RP2Decimal("1") - sold_percent)

                if transaction_cost_basis > ZERO:
                    value = asset_cost_bases.setdefault(asset, ZERO)
                    value += transaction_cost_basis
                    asset_cost_bases[asset] = value

                    total_cost_basis += transaction_cost_basis

            # process balance set data for the asset to collect holder and crypto balance data.
            for balance_set in computed_data.balance_set:
                if balance_set.final_balance > ZERO:
                    if balance_set.holder not in holders:
                        holders.append(balance_set.holder)

                    if asset not in asset_crypto_balance_holder:
                        asset_crypto_balance_holder[asset] = {}
                        asset_crypto_balance_holder_exchange[asset] = {}

                    if balance_set.holder not in asset_crypto_balance_holder[asset]:
                        asset_crypto_balance_holder[asset][balance_set.holder] = ZERO
                        asset_crypto_balance_holder_exchange[asset][balance_set.holder] = {}

                    asset_crypto_balance_holder[asset][balance_set.holder] += balance_set.final_balance

                    if balance_set.exchange not in asset_crypto_balance_holder_exchange[asset][balance_set.holder]:
                        asset_crypto_balance_holder_exchange[asset][balance_set.holder][balance_set.exchange] = balance_set.final_balance

        # Now looping through the assets to do the reporting.
        for asset, asset_cost_basis in asset_cost_bases.items():
            total_crypto_balance = ZERO
            for crypto_balance in asset_crypto_balance_holder[asset].values():
                total_crypto_balance += crypto_balance

            unit_cost_basis: RP2Decimal = asset_cost_basis / total_crypto_balance

            # For report clarity, change how much precision is displayed in the output based on the unit price. The raw value is
            # included in the output, so if the user desires they can change the cell format in the resulting file.
            # The default windowing is set up to hopefully give a good user experience for high value cryptos like BTC
            # all the way through cryptos with minute unit values like SHIB.
            unit_data_style: str = "fiat"
            if _FIAT_UNIT_DATA_STYLE_4_DECIMAL_MINIMUM <= unit_cost_basis < _FIAT_UNIT_DATA_STYLE_2_DECIMAL_MINIMUM:
                unit_data_style = "fiat_unit_4"
            elif unit_cost_basis < _FIAT_UNIT_DATA_STYLE_4_DECIMAL_MINIMUM:
                unit_data_style = "fiat_unit_7"

            # Add this asset to the Input sheet where the user will enter pricing value for the calculations
            input_sheet.append_rows(1)
            input_row_index: int = row_indexes[_INPUT]
            self._fill_cell(input_sheet, input_row_index, 0, asset)
            self._fill_cell(input_sheet, input_row_index, 1, _INPUT_VALUE_STRING, data_style="fiat_unit_7")
            row_indexes[_INPUT] = input_row_index + 1

            _vlookup_formula: str = ""
            _lookup_field: str = ""

            # Complete the asset table.
            for holder, holder_crypto_balance in asset_crypto_balance_holder[asset].items():
                holder_cost_basis: RP2Decimal = holder_crypto_balance * unit_cost_basis

                asset_sheet.append_rows(1)
                asset_row_index: int = row_indexes[_ASSET]
                _vlookup_formula = f"VLOOKUP(A{asset_row_index+1};${_('Input')}.A:B;2;0)"
                _lookup_field = f'=IF({_vlookup_formula}="{_INPUT_VALUE_STRING}";"{_REPORT_INPUT_VALUE_STRING}";{_vlookup_formula}'

                self._fill_cell(asset_sheet, asset_row_index, 0, asset)
                self._fill_cell(asset_sheet, asset_row_index, 1, holder)
                self._fill_cell(asset_sheet, asset_row_index, 2, holder_crypto_balance, data_style="crypto")
                self._fill_cell(asset_sheet, asset_row_index, 3, unit_cost_basis, data_style=unit_data_style)
                self._fill_cell(asset_sheet, asset_row_index, 4, holder_cost_basis, data_style="fiat")
                self._fill_cell(asset_sheet, asset_row_index, 5, holder_cost_basis / total_cost_basis, data_style="percent")
                self._fill_cell(asset_sheet, asset_row_index, 6, _lookup_field, data_style=unit_data_style)
                self._fill_cell(asset_sheet, asset_row_index, 7, f"=C{asset_row_index+1}*G{asset_row_index+1}", data_style="fiat")
                self._fill_cell(asset_sheet, asset_row_index, 8, f"=H{asset_row_index+1}-E{asset_row_index+1}", data_style="fiat")
                self._fill_cell(asset_sheet, asset_row_index, 9, f"=(H{asset_row_index+1}-E{asset_row_index+1})/E{asset_row_index+1}", data_style="percent")
                row_indexes[_ASSET] = asset_row_index + 1

            # Generate the Asset/Exchange table which will calc vals that will feed the asset table.
            for holder, exchanges in asset_crypto_balance_holder_exchange[asset].items():
                for exchange, crypto_exchange_balance in exchanges.items():
                    exchange_cost_basis: RP2Decimal = crypto_exchange_balance * unit_cost_basis

                    asset_exchange_sheet.append_rows(1)
                    asset_exchange_row_index: int = row_indexes[_ASSET_EXCHANGE]
                    _vlookup_formula = f"VLOOKUP(A{asset_exchange_row_index+1};${_('Input')}.A:B;2;0)"
                    _lookup_field = f'=IF({_vlookup_formula}="{_INPUT_VALUE_STRING}";"{_REPORT_INPUT_VALUE_STRING}";{_vlookup_formula}'

                    self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 0, asset)
                    self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 1, holder)
                    self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 2, exchange)
                    self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 3, crypto_exchange_balance, data_style="crypto")
                    self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 4, unit_cost_basis, data_style=unit_data_style)
                    self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 5, exchange_cost_basis, data_style="fiat")
                    self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 6, exchange_cost_basis / total_cost_basis, data_style="percent")
                    self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 7, _lookup_field, data_style=unit_data_style)
                    self._fill_cell(
                        asset_exchange_sheet, asset_exchange_row_index, 8, f"=D{asset_exchange_row_index+1}*H{asset_exchange_row_index+1}", data_style="fiat"
                    )
                    self._fill_cell(
                        asset_exchange_sheet, asset_exchange_row_index, 9, f"=I{asset_exchange_row_index+1}-F{asset_exchange_row_index+1}", data_style="fiat"
                    )
                    self._fill_cell(
                        asset_exchange_sheet,
                        asset_exchange_row_index,
                        10,
                        f"=(I{asset_exchange_row_index+1}-F{asset_exchange_row_index+1})/F{asset_exchange_row_index+1}",
                        data_style="percent",
                    )
                    row_indexes[_ASSET_EXCHANGE] = asset_exchange_row_index + 1

        # There are several portfolio-wide fields in the output that are dependent on values the user enters into the Input tab for
        # live calculation that cannot be accounted for in this report. Since I want to include a totals row in the output, I cannot do
        # full-column sums, e.g. =G4/SUM(G:G), so instead I am using the row index and the header rows info to scope the SUM to the
        # actual rows I know I have placed data into.

        asset_row_index = row_indexes[_ASSET]
        for row_idx in range(self.HEADER_ROWS, row_indexes[_ASSET]):
            self._fill_cell(asset_sheet, row_idx, 10, f"=I{row_idx+1}/SUM(E${self.HEADER_ROWS+1}:E${asset_row_index})", data_style="percent")
            self._fill_cell(asset_sheet, row_idx, 11, f"=H{row_idx+1}/SUM(H${self.HEADER_ROWS+1}:H${asset_row_index})", data_style="percent")

        asset_exchange_row_index = row_indexes[_ASSET_EXCHANGE]
        for row_idx in range(self.HEADER_ROWS, row_indexes[_ASSET_EXCHANGE]):
            self._fill_cell(asset_exchange_sheet, row_idx, 11, f"=J{row_idx+1}/SUM(F${self.HEADER_ROWS+1}:F${asset_exchange_row_index})", data_style="percent")
            self._fill_cell(asset_exchange_sheet, row_idx, 12, f"=I{row_idx+1}/SUM(I${self.HEADER_ROWS+1}:I${asset_exchange_row_index})", data_style="percent")

        # Save the last row index containing data so multiple total rows can be added.
        last_data_row_indexes = row_indexes.copy()

        # Asset sheet totals.
        if len(holders) > 1:
            for holder in holders:
                asset_sheet.append_rows(1)
                asset_row_index = row_indexes[_ASSET]
                last_data_index = last_data_row_indexes[_ASSET]
                self._fill_cell(asset_sheet, asset_row_index, 0, _("Total"), visual_style="bold_border")
                self._fill_cell(asset_sheet, asset_row_index, 1, holder, visual_style="bold_border")
                self._fill_cell(asset_sheet, asset_row_index, 2, "", visual_style="bold_border")
                self._fill_cell(asset_sheet, asset_row_index, 3, "", visual_style="bold_border")
                self._fill_cell(
                    asset_sheet,
                    asset_row_index,
                    4,
                    f'=SUMIF(B${self.HEADER_ROWS+1}:B${last_data_index};"{holder}";E${self.HEADER_ROWS+1}:E${last_data_index})',
                    visual_style="bold_border",
                    data_style="fiat",
                )
                self._fill_cell(asset_sheet, asset_row_index, 5, "", visual_style="bold_border")
                self._fill_cell(asset_sheet, asset_row_index, 6, "", visual_style="bold_border")
                self._fill_cell(
                    asset_sheet,
                    asset_row_index,
                    7,
                    f'=SUMIF(B${self.HEADER_ROWS+1}:B${last_data_index};"{holder}";H${self.HEADER_ROWS+1}:H${last_data_index})',
                    visual_style="bold_border",
                    data_style="fiat",
                )
                self._fill_cell(
                    asset_sheet,
                    asset_row_index,
                    8,
                    f'=SUMIF(B${self.HEADER_ROWS+1}:B${last_data_index};"{holder}";I${self.HEADER_ROWS+1}:I${last_data_index})',
                    visual_style="bold_border",
                    data_style="fiat",
                )
                self._fill_cell(
                    asset_sheet,
                    asset_row_index,
                    9,
                    f"=(H{asset_row_index+1}-E{asset_row_index+1})/E{asset_row_index+1}",
                    visual_style="bold_border",
                    data_style="percent",
                )
                self._fill_cell(asset_sheet, asset_row_index, 10, "", visual_style="bold_border")
                self._fill_cell(asset_sheet, asset_row_index, 11, "", visual_style="bold_border")
                row_indexes[_ASSET] = asset_row_index + 1

        asset_sheet.append_rows(1)
        asset_row_index = row_indexes[_ASSET]
        last_data_index = last_data_row_indexes[_ASSET]
        self._fill_cell(asset_sheet, asset_row_index, 0, _("Grand Total"), visual_style="bold_border")
        self._fill_cell(asset_sheet, asset_row_index, 1, "", visual_style="bold_border")
        self._fill_cell(asset_sheet, asset_row_index, 2, "", visual_style="bold_border")
        self._fill_cell(asset_sheet, asset_row_index, 3, "", visual_style="bold_border")
        self._fill_cell(asset_sheet, asset_row_index, 4, f"=SUM(E${self.HEADER_ROWS+1}:E${last_data_index})", visual_style="bold_border", data_style="fiat")
        self._fill_cell(asset_sheet, asset_row_index, 5, "", visual_style="bold_border")
        self._fill_cell(asset_sheet, asset_row_index, 6, "", visual_style="bold_border")
        self._fill_cell(asset_sheet, asset_row_index, 7, f"=SUM(H${self.HEADER_ROWS+1}:H${last_data_index})", visual_style="bold_border", data_style="fiat")
        self._fill_cell(asset_sheet, asset_row_index, 8, f"=SUM(I${self.HEADER_ROWS+1}:I${last_data_index})", visual_style="bold_border", data_style="fiat")
        self._fill_cell(
            asset_sheet,
            asset_row_index,
            9,
            f"=(H{asset_row_index+1}-E{asset_row_index+1})/E{asset_row_index+1}",
            visual_style="bold_border",
            data_style="percent",
        )
        self._fill_cell(asset_sheet, asset_row_index, 10, "", visual_style="bold_border")
        self._fill_cell(asset_sheet, asset_row_index, 11, "", visual_style="bold_border")
        row_indexes[_ASSET] = asset_row_index + 1

        # Asset - Exchange sheet totals.
        if len(holders) > 1:
            for holder in holders:
                asset_exchange_sheet.append_rows(1)
                asset_exchange_row_index = row_indexes[_ASSET_EXCHANGE]
                last_data_index = last_data_row_indexes[_ASSET_EXCHANGE]
                self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 0, _("Total"), visual_style="bold_border")
                self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 1, holder, visual_style="bold_border")
                self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 2, "", visual_style="bold_border")
                self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 3, "", visual_style="bold_border")
                self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 4, "", visual_style="bold_border")
                self._fill_cell(
                    asset_exchange_sheet,
                    asset_exchange_row_index,
                    5,
                    f'=SUMIF(B${self.HEADER_ROWS+1}:B${last_data_index};"{holder}";F${self.HEADER_ROWS+1}:F${last_data_index})',
                    visual_style="bold_border",
                    data_style="fiat",
                )
                self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 6, "", visual_style="bold_border")
                self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 7, "", visual_style="bold_border")
                self._fill_cell(
                    asset_exchange_sheet,
                    asset_exchange_row_index,
                    8,
                    f'=SUMIF(B${self.HEADER_ROWS+1}:B${last_data_index};"{holder}";I${self.HEADER_ROWS+1}:I${last_data_index})',
                    visual_style="bold_border",
                    data_style="fiat",
                )
                self._fill_cell(
                    asset_exchange_sheet,
                    asset_exchange_row_index,
                    9,
                    f'=SUMIF(B${self.HEADER_ROWS+1}:B${last_data_index};"{holder}";J${self.HEADER_ROWS+1}:J${last_data_index})',
                    visual_style="bold_border",
                    data_style="fiat",
                )
                self._fill_cell(
                    asset_exchange_sheet,
                    asset_exchange_row_index,
                    10,
                    f"=(I{asset_exchange_row_index+1}-F{asset_exchange_row_index+1})/F{asset_exchange_row_index+1}",
                    visual_style="bold_border",
                    data_style="percent",
                )
                self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 11, "", visual_style="bold_border")
                self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 12, "", visual_style="bold_border")
                row_indexes[_ASSET_EXCHANGE] = asset_exchange_row_index + 1

        asset_exchange_sheet.append_rows(1)
        asset_exchange_row_index = row_indexes[_ASSET_EXCHANGE]
        last_data_index = last_data_row_indexes[_ASSET_EXCHANGE]
        self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 0, _("Grand Total"), visual_style="bold_border")
        self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 1, "", visual_style="bold_border")
        self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 2, "", visual_style="bold_border")
        self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 3, "", visual_style="bold_border")
        self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 4, "", visual_style="bold_border")
        self._fill_cell(
            asset_exchange_sheet,
            asset_exchange_row_index,
            5,
            f"=SUM(F${self.HEADER_ROWS+1}:F${last_data_index})",
            visual_style="bold_border",
            data_style="fiat",
        )
        self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 6, "", visual_style="bold_border")
        self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 7, "", visual_style="bold_border")
        self._fill_cell(
            asset_exchange_sheet,
            asset_exchange_row_index,
            8,
            f"=SUM(I${self.HEADER_ROWS+1}:I${last_data_index})",
            visual_style="bold_border",
            data_style="fiat",
        )
        self._fill_cell(
            asset_exchange_sheet,
            asset_exchange_row_index,
            9,
            f"=SUM(J${self.HEADER_ROWS+1}:J${last_data_index})",
            visual_style="bold_border",
            data_style="fiat",
        )
        self._fill_cell(
            asset_exchange_sheet,
            asset_exchange_row_index,
            10,
            f"=(I{asset_exchange_row_index+1}-F{asset_exchange_row_index+1})/F{asset_exchange_row_index+1}",
            visual_style="bold_border",
            data_style="percent",
        )
        self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 11, "", visual_style="bold_border")
        self._fill_cell(asset_exchange_sheet, asset_exchange_row_index, 12, "", visual_style="bold_border")
        row_indexes[_ASSET_EXCHANGE] = asset_exchange_row_index + 1

        asset_sheet.name = _("Asset")
        asset_exchange_sheet.name = _("Asset - Exchange")
        input_sheet.name = _("Input")

        output_file.save()
        LOGGER.info("Plugin '%s' output: %s", __name__, Path(output_file.docname).resolve())



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from configparser import ConfigParser, SectionProxy
from datetime import date, datetime
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Set

from dateutil.parser import parse
from jsonschema import validate

from rp2.abstract_country import AbstractCountry
from rp2.configuration_schema import CONFIGURATION_SCHEMA
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2TypeError, RP2ValueError

MIN_DATE: date = date(1970, 1, 1)
MAX_DATE: date = date(9999, 12, 31)

REPORT_GENERATOR_PACKAGE = "rp2.plugin.report"


class Keyword(Enum):
    ACCOUNTING_METHODS: str = "accounting_methods"
    ASSET: str = "asset"
    ASSETS: str = "assets"
    CRYPTO_FEE: str = "crypto_fee"
    CRYPTO_IN: str = "crypto_in"
    CRYPTO_OUT_NO_FEE: str = "crypto_out_no_fee"
    CRYPTO_OUT_WITH_FEE: str = "crypto_out_with_fee"
    CRYPTO_RECEIVED: str = "crypto_received"
    CRYPTO_SENT: str = "crypto_sent"
    EXCHANGE: str = "exchange"
    EXCHANGES: str = "exchanges"
    FIAT_FEE: str = "fiat_fee"
    FIAT_IN_NO_FEE: str = "fiat_in_no_fee"
    FIAT_IN_WITH_FEE: str = "fiat_in_with_fee"
    FIAT_OUT_NO_FEE: str = "fiat_out_no_fee"
    FROM_EXCHANGE: str = "from_exchange"
    FROM_HOLDER: str = "from_holder"
    GENERAL: str = "general"
    GENERATORS: str = "generators"
    HOLDER: str = "holder"
    HOLDERS: str = "holders"
    IN_HEADER: str = "in_header"
    INTRA_HEADER: str = "intra_header"
    NOTES: str = "notes"
    OUT_HEADER: str = "out_header"
    SPOT_PRICE: str = "spot_price"
    TIMESTAMP: str = "timestamp"
    TO_EXCHANGE: str = "to_exchange"
    TO_HOLDER: str = "to_holder"
    TRANSACTION_TYPE: str = "transaction_type"
    UNIQUE_ID: str = "unique_id"


_HEADER_COLUMNS: Dict[str, Set[str]] = {
    Keyword.IN_HEADER.value: {
        Keyword.TIMESTAMP.value,
        Keyword.ASSET.value,
        Keyword.EXCHANGE.value,
        Keyword.HOLDER.value,
        Keyword.TRANSACTION_TYPE.value,
        Keyword.SPOT_PRICE.value,
        Keyword.CRYPTO_IN.value,
        Keyword.CRYPTO_FEE.value,
        Keyword.FIAT_IN_NO_FEE.value,
        Keyword.FIAT_IN_WITH_FEE.value,
        Keyword.FIAT_FEE.value,
        Keyword.UNIQUE_ID.value,
        Keyword.NOTES.value,
    },
    Keyword.OUT_HEADER.value: {
        Keyword.TIMESTAMP.value,
        Keyword.ASSET.value,
        Keyword.EXCHANGE.value,
        Keyword.HOLDER.value,
        Keyword.TRANSACTION_TYPE.value,
        Keyword.SPOT_PRICE.value,
        Keyword.CRYPTO_OUT_NO_FEE.value,
        Keyword.CRYPTO_FEE.value,
        Keyword.CRYPTO_OUT_WITH_FEE.value,
        Keyword.FIAT_OUT_NO_FEE.value,
        Keyword.FIAT_FEE.value,
        Keyword.UNIQUE_ID.value,
        Keyword.NOTES.value,
    },
    Keyword.INTRA_HEADER.value: {
        Keyword.TIMESTAMP.value,
        Keyword.ASSET.value,
        Keyword.FROM_EXCHANGE.value,
        Keyword.FROM_HOLDER.value,
        Keyword.TO_EXCHANGE.value,
        Keyword.TO_HOLDER.value,
        Keyword.SPOT_PRICE.value,
        Keyword.CRYPTO_SENT.value,
        Keyword.CRYPTO_RECEIVED.value,
        Keyword.UNIQUE_ID.value,
        Keyword.NOTES.value,
    },
}


class Configuration:  # pylint: disable=too-many-public-methods
    @classmethod
    def type_check(cls, name: str, instance: "Configuration") -> "Configuration":
        cls.type_check_parameter_name(name)
        if not isinstance(instance, cls):
            raise RP2TypeError(f"Parameter '{name}' is not of type {cls.__name__}: {instance}")
        return instance

    def __init__(  # pylint: disable=too-many-branches
        self,
        configuration_path: str,
        country: AbstractCountry,
        from_date: date = MIN_DATE,
        to_date: date = MAX_DATE,
        allow_negative_balances: bool = False,
    ) -> None:
        self.__configuration_path: str = self.type_check_string("configuration_path", configuration_path)
        self.__country = AbstractCountry.type_check("country", country)
        if not isinstance(from_date, date):
            raise RP2TypeError("Parameter 'from_date' is not of type date")
        self.__from_date: date = from_date
        if not isinstance(to_date, date):
            raise RP2TypeError("Parameter 'to_date' is not of type date")
        self.__to_date: date = to_date
        if self.__from_date > self.__to_date:
            raise RP2ValueError("Parameter from_date cannot be greater than to_date")
        self.__allow_negative_balances = self.type_check_bool("allow_negative_balances", allow_negative_balances)

        self.__in_header: Dict[str, int] = {}
        self.__out_header: Dict[str, int] = {}
        self.__intra_header: Dict[str, int] = {}
        self.__assets: Set[str] = set()
        self.__exchanges: Set[str] = set()
        self.__holders: Set[str] = set()
        self.__generators: Set[str] = {f"{REPORT_GENERATOR_PACKAGE}.{generator}" for generator in country.get_report_generators()}
        self.__years_2_accounting_method_names: Dict[int, str] = {}

        if not Path(configuration_path).exists():
            raise RP2ValueError(f"Error: {configuration_path} does not exist")

        with open(configuration_path, encoding="utf-8") as configuration_file:
            try:
                # JSON configuration is deprecated: if JSON is detected, raise an exception
                json_configuration: Any = json.load(configuration_file)
                validate(instance=json_configuration, schema=CONFIGURATION_SCHEMA)
                raise RP2ValueError(
                    "Configuration file uses the deprecated JSON-format, instead of the INI format. "
                    f"To convert the JSON configuration to INI use: rp2_config {configuration_path}"
                )
            except json.JSONDecodeError:
                pass

            ini_configuration: ConfigParser = ConfigParser()
            ini_configuration.read(configuration_path)

            for section_name in ini_configuration.sections():
                section_name = section_name.strip()
                normalized_section_name: str = section_name.split(" ", 1)[0].strip()
                if normalized_section_name == Keyword.GENERAL.value:
                    if self.__assets or self.__exchanges or self.__holders:
                        raise RP2ValueError(f"{configuration_path}: section '{normalized_section_name}' found multiple times in configuration file")
                    self.__assets = self._validate_string_set(Keyword.ASSETS.value, ini_configuration[section_name], configuration_path)
                    self.__exchanges = self._validate_string_set(Keyword.EXCHANGES.value, ini_configuration[section_name], configuration_path)
                    self.__holders = self._validate_string_set(Keyword.HOLDERS.value, ini_configuration[section_name], configuration_path)
                    if Keyword.GENERATORS.value in ini_configuration:
                        self.__generators = self._validate_string_set(Keyword.GENERATORS.value, ini_configuration[section_name], configuration_path)
                elif normalized_section_name == Keyword.IN_HEADER.value:
                    if self.__in_header:
                        raise RP2ValueError(f"{configuration_path}: section '{normalized_section_name}' found multiple times in configuration file")
                    self.__in_header = self._validate_header_section(ini_configuration[section_name], normalized_section_name, configuration_path)
                elif normalized_section_name == Keyword.OUT_HEADER.value:
                    if self.__out_header:
                        raise RP2ValueError(f"{configuration_path}: section '{normalized_section_name}' found multiple times in configuration file")
                    self.__out_header = self._validate_header_section(ini_configuration[section_name], normalized_section_name, configuration_path)
                elif normalized_section_name == Keyword.INTRA_HEADER.value:
                    if self.__intra_header:
                        raise RP2ValueError(f"{configuration_path}: section '{normalized_section_name}' found multiple times in configuration file")
                    self.__intra_header = self._validate_header_section(ini_configuration[section_name], normalized_section_name, configuration_path)
                elif normalized_section_name == Keyword.ACCOUNTING_METHODS.value:
                    if self.__years_2_accounting_method_names:
                        raise RP2ValueError(f"{configuration_path}: section '{normalized_section_name}' found multiple times in configuration file")
                    self.__years_2_accounting_method_names = self._validate_accounting_method_section(ini_configuration[section_name], configuration_path)
                else:
                    raise RP2ValueError(f"{configuration_path}: invalid section '{section_name}' found")

        if not self.__assets:
            raise RP2ValueError(f"{configuration_path}: no '{Keyword.ASSETS.value}' field defined in {Keyword.GENERAL.value} section")
        if not self.__exchanges:
            raise RP2ValueError(f"{configuration_path}: no '{Keyword.EXCHANGES.value}' field defined in {Keyword.GENERAL.value} section")
        if not self.__holders:
            raise RP2ValueError(f"{configuration_path}: no '{Keyword.HOLDERS.value}' field defined in {Keyword.GENERAL.value} section")
        if not self.__in_header:
            raise RP2ValueError(f"{configuration_path}: empty '{Keyword.IN_HEADER.value}' section")
        if not self.__out_header:
            raise RP2ValueError(f"{configuration_path}: empty '{Keyword.OUT_HEADER.value}' section")
        if not self.__intra_header:
            raise RP2ValueError(f"{configuration_path}: empty '{Keyword.INTRA_HEADER.value}' section")

        # Used by __repr__()
        self.__sorted_assets: List[str] = sorted(self.__assets)
        self.__sorted_exchanges: List[str] = sorted(self.__exchanges)
        self.__sorted_holders: List[str] = sorted(self.__holders)

    def _validate_string_set(self, field_name: str, section: SectionProxy, configuration_path: str) -> Set[str]:
        if field_name not in section:
            raise RP2ValueError(f"{configuration_path}: section '{section.name}' doesn't contain mandatory field '{field_name}'")
        list_as_string = section[field_name]
        if not list_as_string or not list_as_string.strip():
            raise RP2ValueError(f"{configuration_path}: field '{field_name}' in section '{section.name}' cannot be empty")
        list_as_values: List[str] = [value.strip() for value in list_as_string.split(",")]
        if not list_as_values:
            raise RP2ValueError(f"{configuration_path}: field '{field_name}' in section '{section.name}' cannot be empty")
        result: Set[str] = set()
        for element in list_as_values:
            if not element:
                raise RP2ValueError(f"{configuration_path}: field '{field_name}' in section '{section.name}' cannot contain empty elements")
            if not isinstance(element, str):
                raise RP2ValueError(f"{configuration_path}: field '{field_name}' in section '{section.name}' cannot contain non-string elements")
            if element in result:
                raise RP2ValueError(f"{configuration_path}: field '{field_name}' in section '{section.name}' contains duplicate elements: {element}")
            result.add(element)
        return set(list_as_values)

    def _validate_header_section(self, section: SectionProxy, normalized_section_name: str, configuration_path: str) -> Dict[str, int]:
        if not section:
            raise RP2ValueError(f"{configuration_path}: section '{section.name}' cannot be empty")
        header_2_column: Dict[str, int] = {}
        column_to_header: Dict[int, str] = {}
        for header, column in section.items():
            try:
                column_value: int = int(column.strip())
                if column_value < 0:
                    raise RP2ValueError(
                        f"{configuration_path}: invalid column value for field '{header}' in section '{section.name}' (positive integer was expected): {column}"
                    )
                if column_value in column_to_header:
                    raise RP2ValueError(
                        f"{configuration_path}: fields '{column_to_header[column_value]}' and "
                        f"'{header}' have the same value in section '{section.name}': {column_value}"
                    )
                if header not in _HEADER_COLUMNS[normalized_section_name]:
                    raise RP2ValueError(f"{configuration_path}: invalid column header in section '{section.name}': {header}")
                header_2_column[header.strip()] = column_value
                column_to_header[column_value] = header
            except ValueError as exc:
                raise RP2ValueError(
                    f"{configuration_path}: invalid column value for field '{header}' in section '{section.name}' (integer was expected): {column}"
                ) from exc
            except TypeError as exc:
                raise RP2ValueError(
                    f"{configuration_path}: invalid value type for field '{header}' in section '{section.name}' (integer was expected): {column}"
                ) from exc
        return header_2_column

    def _validate_accounting_method_section(self, section: SectionProxy, configuration_path: str) -> Dict[int, str]:
        if not section:
            raise RP2ValueError(f"{configuration_path}: section '{section.name}' cannot be empty")
        result: Dict[int, str] = {}
        for year, method in section.items():
            try:
                numeric_year: int = int(year.strip())
                if numeric_year < MIN_DATE.year:
                    raise RP2ValueError(
                        f"{configuration_path}: invalid year value in accounting method section (integer > {MIN_DATE.year} was expected): {year}"
                    )
                result[numeric_year] = method.strip()
            except ValueError as exc:
                raise RP2ValueError(f"{configuration_path}: invalid year value in accounting method section (integer was expected): {year}") from exc
            except TypeError as exc:
                raise RP2ValueError(f"{configuration_path}: invalid year type in accounting method section (integer was expected): {year}") from exc
        return result

    def __repr__(self) -> str:
        return (
            f"Configuration(configuration_path={self.configuration_path}, "
            f"country={repr(self.country)}, "
            f"from_date={self.from_date if self.from_date > MIN_DATE else 'non-specified'}, "
            f"to_date={self.to_date if self.to_date < MAX_DATE else 'non-specified'}, "
            f"in_header={self.__in_header}, "
            f"out_header={self.__out_header}, "
            f"intra_header={self.__intra_header}, "
            f"assets={self.__sorted_assets}, "
            f"exchanges={self.__sorted_exchanges}, "
            f"holders={self.__sorted_holders})"
        )

    @property
    def configuration_path(self) -> str:
        return self.__configuration_path

    @property
    def country(self) -> AbstractCountry:
        return self.__country

    @property
    def from_date(self) -> date:
        return self.__from_date

    @property
    def to_date(self) -> date:
        return self.__to_date

    @property
    def allow_negative_balances(self) -> bool:
        return self.__allow_negative_balances

    @property
    def assets(self) -> Set[str]:
        return self.__assets

    @property
    def generators(self) -> Set[str]:
        return self.__generators

    @property
    def years_2_accounting_method_names(self) -> Dict[int, str]:
        return self.__years_2_accounting_method_names

    def __get_table_constructor_argument_pack(self, data: List[Any], table_type: str, header: Dict[str, int]) -> Dict[str, Any]:
        if not isinstance(data, List):
            raise RP2TypeError(f"Parameter 'data' value is not a List: {data}")
        max_column: int = header[max(header, key=header.get)]  # type: ignore
        if len(data) <= max_column:
            raise RP2ValueError(
                f"Parameter 'data' has length {len(data)}, but required minimum from {table_type}-table headers in "
                f"{self.__configuration_path} is {max_column + 1}: {data}"
            )
        pack: Dict[str, Any] = {argument: data[position] for argument, position in header.items()}

        return pack

    def get_in_table_constructor_argument_pack(self, data: List[Any]) -> Dict[str, Any]:
        return self.__get_table_constructor_argument_pack(data, "in", self.__in_header)

    def get_out_table_constructor_argument_pack(self, data: List[Any]) -> Dict[str, Any]:
        return self.__get_table_constructor_argument_pack(data, "out", self.__out_header)

    def get_intra_table_constructor_argument_pack(self, data: List[Any]) -> Dict[str, Any]:
        return self.__get_table_constructor_argument_pack(data, "intra", self.__intra_header)

    def get_in_table_column_position(self, input_parameter: str) -> int:
        self.type_check_string("input_parameter", input_parameter)
        if input_parameter not in self.__in_header:
            raise RP2ValueError(f"Unknown 'input_parameter' {input_parameter}")
        return self.__in_header[input_parameter]

    def get_out_table_column_position(self, input_parameter: str) -> int:
        self.type_check_string("input_parameter", input_parameter)
        if input_parameter not in self.__out_header:
            raise RP2ValueError(f"Unknown 'input_parameter' {input_parameter}")
        return self.__out_header[input_parameter]

    def get_intra_table_column_position(self, input_parameter: str) -> int:
        self.type_check_string("input_parameter", input_parameter)
        if input_parameter not in self.__intra_header:
            raise RP2ValueError(f"Unknown 'input_parameter' {input_parameter}")
        return self.__intra_header[input_parameter]

    @classmethod
    def type_check_internal_id(cls, name: str, value: int) -> int:
        return cls.type_check_int(name, value)

    @classmethod
    def type_check_timestamp_from_string(cls, name: str, value: str) -> datetime:
        cls.type_check_string(name, value)
        try:
            result: datetime = parse(value)
        except Exception as exc:
            raise RP2ValueError(f"Error parsing parameter '{name}': {str(exc)}") from exc
        if result.tzinfo is None:
            raise RP2ValueError(f"Parameter '{name}' value has no timezone info: {value}")
        return result

    def type_check_exchange(self, name: str, value: str) -> str:
        self.type_check_string(name, value)
        if value not in self.__exchanges:
            raise RP2ValueError(f"Parameter '{name}' value is not known: {value}")
        return value

    def type_check_holder(self, name: str, value: str) -> str:
        self.type_check_string(name, value)
        if value not in self.__holders:
            raise RP2ValueError(f"Parameter '{name}' value is not known: {value}")
        return value

    def type_check_asset(self, name: str, value: str) -> str:
        self.type_check_string(name, value)
        if value not in self.__assets:
            raise RP2ValueError(f"Parameter '{name}' value is not known: {value}")
        return value

    @classmethod
    def type_check_parameter_name(cls, name: str) -> str:
        if not isinstance(name, str):
            raise RP2TypeError(f"Parameter name is not a string: {name}")
        return name

    @classmethod
    def type_check_string_or_integer(cls, name: str, value: str) -> str:
        cls.type_check_parameter_name(name)
        if not isinstance(value, (str, int, float)):
            raise RP2TypeError(f"Parameter '{name}' has non-string value {repr(value)}")
        return str(value)

    @classmethod
    def type_check_string(cls, name: str, value: str) -> str:
        cls.type_check_parameter_name(name)
        if not isinstance(value, str):
            raise RP2TypeError(f"Parameter '{name}' has non-string value {repr(value)}")
        return value

    @classmethod
    def type_check_positive_int(cls, name: str, value: int, non_zero: bool = False) -> int:
        result: int = cls.type_check_int(name, value)
        if result < 0:
            raise RP2ValueError(f"Parameter '{name}' has non-positive value {value}")
        if non_zero and result == 0:
            raise RP2ValueError(f"Parameter '{name}' has zero value")
        return result

    @classmethod
    def type_check_int(cls, name: str, value: int) -> int:
        cls.type_check_parameter_name(name)
        if not isinstance(value, int):
            raise RP2TypeError(f"Parameter '{name}' has non-integer value {repr(value)}")
        return value

    @classmethod
    def type_check_positive_float(cls, name: str, value: float, non_zero: bool = False) -> float:
        result: float = cls.type_check_float(name, value)
        if result < 0:
            raise RP2ValueError(f"Parameter '{name}' has non-positive value {value}")
        if non_zero and result == 0:
            raise RP2ValueError(f"Parameter '{name}' has zero value")
        return result

    @classmethod
    def type_check_float(cls, name: str, value: float) -> float:
        cls.type_check_parameter_name(name)
        if not isinstance(value, (int, float)):
            raise RP2TypeError(f"Parameter '{name}' has non-numeric value {repr(value)}")
        return value

    @classmethod
    def type_check_bool(cls, name: str, value: bool) -> bool:
        cls.type_check_parameter_name(name)
        if not isinstance(value, bool):
            raise RP2TypeError(f"Parameter '{name}' has non-bool value {repr(value)}")
        return value

    @classmethod
    def type_check_positive_decimal(cls, name: str, value: RP2Decimal, non_zero: bool = False) -> RP2Decimal:
        result: RP2Decimal = cls.type_check_decimal(name, value)
        if result < ZERO:
            raise RP2ValueError(f"Parameter '{name}' has non-positive value {value}")
        if non_zero and result == ZERO:
            raise RP2ValueError(f"Parameter '{name}' has zero value")
        return result

    @classmethod
    def type_check_decimal(cls, name: str, value: RP2Decimal) -> RP2Decimal:
        cls.type_check_parameter_name(name)
        if not isinstance(value, RP2Decimal):
            raise RP2TypeError(f"Parameter '{name}' has non-RP2Decimal value {repr(value)}")
        return value



# Copyright 2023 Larision
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import unittest
from pathlib import Path

from abstract_test_ods_output_diff import AbstractTestODSOutputDiff, OutputPlugins

ROOT_PATH: Path = Path(os.path.dirname(__file__)).parent.absolute()


class TestODSOutputDiff(AbstractTestODSOutputDiff):  # pylint: disable=too-many-public-methods
    output_dir: Path

    @classmethod
    def setUpClass(cls) -> None:
        cls.output_dir = ROOT_PATH / Path("output") / Path(cls.__module__)

        shutil.rmtree(cls.output_dir, ignore_errors=True)

        AbstractTestODSOutputDiff._generate(
            cls.output_dir,
            test_name="crypto_example",
            config="crypto_example",
            method="fifo",
            country="es",
            generation_language="es",
            allow_negative_balances=True,
        )
        AbstractTestODSOutputDiff._generate(cls.output_dir, test_name="test_data", config="test_data", method="fifo", country="es", generation_language="es")
        AbstractTestODSOutputDiff._generate(
            cls.output_dir, test_name="test_data2", config="test_data", method="fifo", country="es", generation_language="es", allow_negative_balances=True
        )
        AbstractTestODSOutputDiff._generate(cls.output_dir, test_name="test_data3", config="test_data", method="fifo", country="es", generation_language="es")
        AbstractTestODSOutputDiff._generate(cls.output_dir, test_name="test_data4", config="test_data4", method="fifo", country="es", generation_language="es")
        AbstractTestODSOutputDiff._generate(
            cls.output_dir,
            test_name="test_many_year_data",
            config="test_data",
            method="fifo",
            country="es",
            generation_language="es",
            allow_negative_balances=True,
        )

    def setUp(self) -> None:
        self.maxDiff = None  # pylint: disable=invalid-name

    def test_crypto_example_rp2_full_report(self) -> None:
        self._compare(
            output_dir=self.output_dir,
            test_name="crypto_example",
            method="fifo",
            country="es",
            output_plugin=OutputPlugins.RP2_FULL_REPORT,
            generation_language="es",
        )

    def test_test_data_rp2_full_report(self) -> None:
        self._compare(
            output_dir=self.output_dir,
            test_name="test_data",
            method="fifo",
            country="es",
            output_plugin=OutputPlugins.RP2_FULL_REPORT,
            generation_language="es",
        )

    def test_test_data2_rp2_full_report(self) -> None:
        self._compare(
            output_dir=self.output_dir,
            test_name="test_data2",
            method="fifo",
            country="es",
            output_plugin=OutputPlugins.RP2_FULL_REPORT,
            generation_language="es",
        )

    def test_test_data3_rp2_full_report(self) -> None:
        self._compare(
            output_dir=self.output_dir,
            test_name="test_data3",
            method="fifo",
            country="es",
            output_plugin=OutputPlugins.RP2_FULL_REPORT,
            generation_language="es",
        )

    def test_test_data4_rp2_full_report(self) -> None:
        self._compare(
            output_dir=self.output_dir,
            test_name="test_data4",
            method="fifo",
            country="es",
            output_plugin=OutputPlugins.RP2_FULL_REPORT,
            generation_language="es",
        )

    def test_test_many_year_data_rp2_full_report(self) -> None:
        self._compare(
            output_dir=self.output_dir,
            test_name="test_many_year_data",
            method="fifo",
            country="es",
            output_plugin=OutputPlugins.RP2_FULL_REPORT,
            generation_language="es",
        )


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, List, Optional

from rp2.abstract_transaction import AbstractTransaction
from rp2.configuration import Configuration
from rp2.entry_types import TransactionType
from rp2.logger import LOGGER
from rp2.rp2_decimal import FIAT_DECIMAL_MASK, ZERO, RP2Decimal
from rp2.rp2_error import RP2TypeError, RP2ValueError


# pylint: disable=too-many-branches
class OutTransaction(AbstractTransaction):
    def __init__(
        self,
        configuration: Configuration,
        timestamp: str,
        asset: str,
        exchange: str,
        holder: str,
        transaction_type: str,
        spot_price: RP2Decimal,
        crypto_out_no_fee: RP2Decimal,
        crypto_fee: RP2Decimal,
        crypto_out_with_fee: Optional[RP2Decimal] = None,
        fiat_out_no_fee: Optional[RP2Decimal] = None,
        fiat_fee: Optional[RP2Decimal] = None,
        internal_id: Optional[int] = None,
        unique_id: Optional[str] = None,
        notes: Optional[str] = None,
    ) -> None:
        super().__init__(configuration, timestamp, asset, transaction_type, spot_price, internal_id, unique_id, notes)

        self.__exchange: str = configuration.type_check_exchange("exchange", exchange)
        self.__holder: str = configuration.type_check_holder("holder", holder)
        self.__crypto_out_no_fee: RP2Decimal
        self.__crypto_fee: RP2Decimal
        self.__fiat_out_with_fee: RP2Decimal
        self.__fiat_out_no_fee: RP2Decimal

        if self.transaction_type == TransactionType.FEE:
            self.__crypto_out_no_fee = configuration.type_check_positive_decimal("crypto_out_no_fee", crypto_out_no_fee)
            if self.__crypto_out_no_fee != ZERO:
                raise RP2ValueError(
                    f"{self.asset} {type(self).__name__} ({self.timestamp}, id {self.internal_id}): fee-typed transaction has non-zero 'crypto_out_no_fee'"
                )
            self.__crypto_fee = configuration.type_check_positive_decimal("crypto_fee", crypto_fee, non_zero=True)
        else:
            if spot_price == ZERO:
                raise RP2ValueError(f"{self.asset} {type(self).__name__} ({self.timestamp}, id {self.internal_id}): parameter 'spot_price' cannot be 0")
            self.__crypto_out_no_fee = configuration.type_check_positive_decimal("crypto_out_no_fee", crypto_out_no_fee, non_zero=True)
            self.__crypto_fee = configuration.type_check_positive_decimal("crypto_fee", crypto_fee)

        # Crypto out with fee is optional. It can be derived from crypto out (no fee) and crypto fee, however some exchanges
        # provide it anyway. If it is provided use it as given by the exchange, if not compute it.
        if crypto_out_with_fee is None:
            self.__crypto_out_with_fee = self.__crypto_out_no_fee + self.crypto_fee
        else:
            self.__crypto_out_with_fee = configuration.type_check_positive_decimal("crypto_out_with_fee", crypto_out_with_fee, non_zero=True)

        # Fiat out without fee and fiat fee are optional. They can be derived from crypto out (no fee), spot price and crypto fee,
        # however some exchanges provide them anyway. If they are provided use them as given by the exchange, if not compute them.
        if fiat_out_no_fee is None:
            self.__fiat_out_no_fee = self.__crypto_out_no_fee * self.spot_price
        else:
            self.__fiat_out_no_fee = configuration.type_check_positive_decimal("fiat_out_no_fee", fiat_out_no_fee, non_zero=True)
        if fiat_fee is None:
            self.__fiat_fee = self.__crypto_fee * self.spot_price
        else:
            self.__fiat_fee = configuration.type_check_positive_decimal("fiat_fee", fiat_fee)
        self.__fiat_out_with_fee = self.__fiat_out_no_fee + self.__fiat_fee

        if self.transaction_type not in (TransactionType.DONATE, TransactionType.FEE, TransactionType.GIFT, TransactionType.SELL):
            raise RP2ValueError(
                f"{self.asset} {type(self).__name__} ({self.timestamp}, id {self.internal_id}): invalid transaction type {self.transaction_type}"
            )

        # If the values provided by the exchange doesn't match the computed one, log a warning.
        if not RP2Decimal.is_equal_within_precision(self.__crypto_out_with_fee, self.__crypto_out_no_fee + self.__crypto_fee, FIAT_DECIMAL_MASK):
            LOGGER.warning(
                "%s %s (%s, id %s): crypto_out_with_fee != crypto_out_no_fee + crypto_fee: %f != %f",
                self.asset,
                type(self).__name__,
                self.timestamp,
                self.internal_id,
                self.__crypto_out_with_fee,
                self.__crypto_out_no_fee + self.__crypto_fee,
            )

        if not RP2Decimal.is_equal_within_precision(self.__crypto_fee * self.spot_price, self.__fiat_fee, FIAT_DECIMAL_MASK):
            LOGGER.warning(
                "%s %s (%s, id %s): crypto_fee * spot_price != fiat_fee: %f != %f",
                self.asset,
                type(self).__name__,
                self.timestamp,
                self.internal_id,
                self.__crypto_fee * self.spot_price,
                self.__fiat_fee,
            )

        if not RP2Decimal.is_equal_within_precision(self.__crypto_out_no_fee * self.spot_price, self.__fiat_out_no_fee, FIAT_DECIMAL_MASK):
            LOGGER.warning(
                "%s %s (%s, id %s): crypto_out_no_fee * spot_price != fiat_out_no_fee: %f != %f",
                self.asset,
                type(self).__name__,
                self.timestamp,
                self.internal_id,
                self.__crypto_out_no_fee * self.spot_price,
                self.__fiat_out_no_fee,
            )

    def to_string(self, indent: int = 0, repr_format: bool = True, extra_data: Optional[List[str]] = None) -> str:
        self.configuration.type_check_positive_int("indent", indent)
        self.configuration.type_check_bool("repr_format", repr_format)
        if extra_data and not isinstance(extra_data, List):
            raise RP2TypeError(f"Parameter 'extra_data' is not of type List: {extra_data}")

        class_specific_data: List[str] = []
        stringify: Callable[[object], str] = repr
        if not repr_format:
            stringify = str
        class_specific_data = [
            f"exchange={stringify(self.exchange)}",
            f"holder={stringify(self.holder)}",
            f"transaction_type={stringify(self.transaction_type)}",
            f"spot_price={self.spot_price:.4f}",
            f"crypto_out_no_fee={self.crypto_out_no_fee:.8f}",
            f"crypto_fee={self.crypto_fee:.8f}",
            f"unique_id={self.unique_id}",
            f"is_taxable={stringify(self.is_taxable())}",
            f"fiat_taxable_amount={self.fiat_taxable_amount:.4f}",
        ]
        if extra_data:
            class_specific_data.extend(extra_data)

        return super().to_string(indent=indent, repr_format=repr_format, extra_data=class_specific_data)

    @property
    def exchange(self) -> str:
        return self.__exchange

    @property
    def holder(self) -> str:
        return self.__holder

    @property
    def crypto_out_no_fee(self) -> RP2Decimal:
        return self.__crypto_out_no_fee

    @property
    def crypto_out_with_fee(self) -> RP2Decimal:
        return self.__crypto_out_with_fee

    @property
    def crypto_fee(self) -> RP2Decimal:
        return self.__crypto_fee

    @property
    def fiat_out_no_fee(self) -> RP2Decimal:
        return self.__fiat_out_no_fee

    @property
    def fiat_out_with_fee(self) -> RP2Decimal:
        return self.__fiat_out_with_fee

    @property
    def fiat_fee(self) -> RP2Decimal:
        return self.__fiat_fee

    # IRS Publication 544 (https://www.irs.gov/publications/p544) explains that sale expenses are deducted from the sale price
    # (see "Example 1" in the "Gain or Loss From Sales and Exchanges" section). A less formal explanation:
    # https://taxbit.com/cryptocurrency-tax-guide. Therefore the fee is considered a deduction and the outgoing amount is not.
    @property
    def crypto_taxable_amount(self) -> RP2Decimal:
        if self.transaction_type == TransactionType.FEE:
            return self.crypto_fee
        return self.crypto_out_no_fee

    @property
    def fiat_taxable_amount(self) -> RP2Decimal:
        if self.transaction_type == TransactionType.FEE:
            return self.fiat_fee
        return self.fiat_out_no_fee

    @property
    def crypto_deduction(self) -> RP2Decimal:
        return self.crypto_fee

    @property
    def fiat_deduction(self) -> RP2Decimal:
        return self.fiat_fee

    @property
    def crypto_balance_change(self) -> RP2Decimal:
        return self.crypto_out_with_fee

    @property
    def fiat_balance_change(self) -> RP2Decimal:
        return self.fiat_out_with_fee

    def is_taxable(self) -> bool:
        return True



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from enum import Enum
from typing import Dict, List, NamedTuple, Optional

from rp2.abstract_transaction import AbstractTransaction
from rp2.in_transaction import InTransaction
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2RuntimeError


class AcquiredLotAndAmount(NamedTuple):
    acquired_lot: InTransaction
    amount: RP2Decimal


class AcquiredLotCandidatesOrder(Enum):
    OLDER_TO_NEWER: str = "older_to_newer"
    NEWER_TO_OLDER: str = "newer_to_older"


class AcquiredLotCandidates:
    def __init__(
        self,
        accounting_method: "AbstractAccountingMethod",
        acquired_lot_list: List[InTransaction],
        acquired_lot_2_partial_amount: Dict[InTransaction, RP2Decimal],
        up_to_index: int,
    ) -> None:
        self.__accounting_method: AbstractAccountingMethod = accounting_method
        self.__acquired_lot_list = acquired_lot_list
        self.__acquired_lot_2_partial_amount = acquired_lot_2_partial_amount
        self.__up_to_index = up_to_index

    def has_partial_amount(self, acquired_lot: InTransaction) -> bool:
        return acquired_lot in self.__acquired_lot_2_partial_amount

    def get_partial_amount(self, acquired_lot: InTransaction) -> RP2Decimal:
        if not self.has_partial_amount(acquired_lot):
            raise RP2RuntimeError(f"Internal error: acquired lot has no partial amount: {acquired_lot}")
        return self.__acquired_lot_2_partial_amount[acquired_lot]

    def set_partial_amount(self, acquired_lot: InTransaction, amount: RP2Decimal) -> None:
        self.__acquired_lot_2_partial_amount[acquired_lot] = amount

    def clear_partial_amount(self, acquired_lot: InTransaction) -> None:
        self.set_partial_amount(acquired_lot, ZERO)

    def __iter__(self) -> "AccountingMethodIterator":
        return AccountingMethodIterator(self.__acquired_lot_list, self.__up_to_index, self.__accounting_method.lot_candidates_order())


class AccountingMethodIterator:
    def __init__(self, acquired_lot_list: List[InTransaction], up_to_index: int, order_type: AcquiredLotCandidatesOrder) -> None:
        self.__acquired_lot_list = acquired_lot_list
        self.__start_index = 0 if order_type == AcquiredLotCandidatesOrder.OLDER_TO_NEWER else up_to_index
        self.__end_index = up_to_index if order_type == AcquiredLotCandidatesOrder.OLDER_TO_NEWER else 0
        self.__step = 1 if order_type == AcquiredLotCandidatesOrder.OLDER_TO_NEWER else -1
        self.__index = self.__start_index
        self.__order_type = order_type

    def _check_index(self) -> bool:
        if self.__order_type == AcquiredLotCandidatesOrder.OLDER_TO_NEWER:
            return self.__index <= self.__end_index
        return self.__index >= self.__end_index

    def __next__(self) -> InTransaction:
        result: Optional[InTransaction] = None
        while self._check_index():
            result = self.__acquired_lot_list[self.__index]
            self.__index += self.__step
            return result
        raise StopIteration(self)


class AbstractAccountingMethod:
    def seek_non_exhausted_acquired_lot(
        self,
        lot_candidates: AcquiredLotCandidates,
        taxable_event: Optional[AbstractTransaction],
        taxable_event_amount: RP2Decimal,
    ) -> Optional[AcquiredLotAndAmount]:
        raise NotImplementedError("Abstract function")

    def lot_candidates_order(self) -> AcquiredLotCandidatesOrder:
        raise NotImplementedError("Abstract function")

    @property
    def name(self) -> str:
        return f"{self.__class__.__module__.rsplit('.', 1)[1]}"

    def __repr__(self) -> str:
        return self.name



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import date
from typing import Dict, List, Optional, cast

from rp2.abstract_entry import AbstractEntry
from rp2.abstract_entry_set import AbstractEntrySet
from rp2.abstract_transaction import AbstractTransaction
from rp2.configuration import MAX_DATE, MIN_DATE, Configuration
from rp2.entry_types import TransactionType
from rp2.gain_loss import GainLoss
from rp2.in_transaction import InTransaction
from rp2.logger import LOGGER
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2RuntimeError, RP2TypeError, RP2ValueError


class GainLossSet(AbstractEntrySet):
    @classmethod
    def type_check(cls, name: str, instance: "GainLossSet") -> "GainLossSet":
        Configuration.type_check_parameter_name(name)
        if not isinstance(instance, cls):
            raise RP2TypeError(f"Parameter '{name}' is not of type {cls.__name__}: {instance}")
        return instance

    def __init__(
        self,
        configuration: Configuration,
        asset: str,
        from_date: date = MIN_DATE,
        to_date: date = MAX_DATE,
    ) -> None:
        super().__init__(configuration, "MIXED", asset, from_date, to_date)
        self.__taxable_events_to_fraction: Dict[GainLoss, int] = {}
        self.__acquired_lots_to_fraction: Dict[GainLoss, int] = {}
        self.__taxable_events_to_number_of_fractions: Dict[AbstractTransaction, int] = {}
        self.__acquired_lots_to_number_of_fractions: Dict[InTransaction, int] = {}
        self.__transaction_type_2_count: Dict[TransactionType, int] = {transaction_type: 0 for transaction_type in TransactionType}

    def add_entry(self, entry: AbstractEntry) -> None:
        GainLoss.type_check("entry", entry)
        super().add_entry(entry)

    def get_transaction_type_count(self, transaction_type: TransactionType) -> int:
        TransactionType.type_check("transaction_type", transaction_type)
        self._check_sort()
        return self.__transaction_type_2_count[transaction_type]

    def get_taxable_event_fraction(self, entry: GainLoss) -> int:
        self._validate_entry(entry)
        self._check_sort()
        return self.__taxable_events_to_fraction[entry]

    def get_acquired_lot_fraction(self, entry: GainLoss) -> int:
        self._validate_entry(entry)
        self._check_sort()
        return self.__acquired_lots_to_fraction[entry]

    def get_taxable_event_number_of_fractions(self, transaction: AbstractTransaction) -> int:
        AbstractTransaction.type_check("transaction", transaction)
        if transaction not in self.__taxable_events_to_number_of_fractions:
            raise RP2ValueError(f"Unknown transaction:\n{transaction}")
        self._check_sort()
        return self.__taxable_events_to_number_of_fractions[transaction]

    def get_acquired_lot_number_of_fractions(self, transaction: InTransaction) -> int:
        InTransaction.type_check("transaction", transaction)
        if transaction not in self.__acquired_lots_to_number_of_fractions:
            raise RP2ValueError(f"Unknown transaction:\n{transaction}")
        self._check_sort()
        return self.__acquired_lots_to_number_of_fractions[transaction]

    def _validate_entry(self, entry: AbstractEntry) -> None:
        GainLoss.type_check("entry", entry)
        super()._validate_entry(entry)

    def _sort_entries(self) -> None:  # pylint: disable=too-many-branches
        LOGGER.debug("Sort Gain-Loss Set:")
        super()._sort_entries()
        entry: AbstractEntry
        gain_loss: Optional[GainLoss] = None
        # Taxable events are always monotonic over time (sorted by ascending date), so we just need scalars to keep
        # track of amount and fraction (see also acquired-lot comment below). On the other hand acquired lots are not always
        # monotonic over time (they can be in any order, depending on the accounting method), so we need dictionaries
        # to keep track of amount and fraction for each lot.
        current_taxable_event_amount: RP2Decimal = ZERO
        current_taxable_event_fraction: int = 0
        current_acquired_lot_amount: Dict[InTransaction, RP2Decimal] = {}
        current_acquired_lot_fraction: Dict[InTransaction, int] = {}

        last_gain_loss_with_acquired_lot: Optional[GainLoss] = None

        # Reset fields that are recomputed at sort time
        self.__taxable_events_to_fraction = {}
        self.__taxable_events_to_number_of_fractions = {}
        self.__acquired_lots_to_fraction = {}
        self.__acquired_lots_to_number_of_fractions = {}
        self.__transaction_type_2_count = {transaction_type: 0 for transaction_type in TransactionType}

        for entry in self._entry_list:
            gain_loss = cast(GainLoss, entry)

            # We're not using the iterator to avoid infinite recursion (we're looping over
            # _entry_list directly), so we need to check time filters manually: stop after
            # to_date so that number of fractions is not affected by lots outside the time filter
            if gain_loss.timestamp.date() > self.to_date:
                break

            count: int = self.__transaction_type_2_count[gain_loss.taxable_event.transaction_type]
            self.__transaction_type_2_count[gain_loss.taxable_event.transaction_type] = count + 1

            if gain_loss.acquired_lot:
                last_gain_loss_with_acquired_lot = gain_loss

            current_taxable_event_amount += gain_loss.crypto_amount
            self.__taxable_events_to_fraction[gain_loss] = current_taxable_event_fraction
            if current_taxable_event_amount == gain_loss.taxable_event.crypto_balance_change:
                # Expected amount reached: reset both fraction and amount
                if gain_loss.taxable_event in self.__taxable_events_to_number_of_fractions:
                    raise RP2ValueError(f"Taxable event crypto amount already exhausted for {gain_loss.taxable_event}")
                self.__taxable_events_to_number_of_fractions[gain_loss.taxable_event] = current_taxable_event_fraction + 1
                LOGGER.debug(
                    "%s (%d - %d): current amount == taxable event (%.16f)",
                    gain_loss.internal_id,
                    current_acquired_lot_fraction[gain_loss.acquired_lot] if gain_loss.acquired_lot in current_acquired_lot_fraction else 0,
                    current_taxable_event_fraction,
                    current_taxable_event_amount,
                )
                current_taxable_event_fraction = 0
                current_taxable_event_amount = ZERO
            elif current_taxable_event_amount < gain_loss.taxable_event.crypto_balance_change:
                LOGGER.debug(
                    "%s (%d - %d): current amount < taxable event (%.16f < %.16f)",
                    gain_loss.internal_id,
                    current_acquired_lot_fraction[gain_loss.acquired_lot] if gain_loss.acquired_lot in current_acquired_lot_fraction else 0,
                    current_taxable_event_fraction,
                    current_taxable_event_amount,
                    gain_loss.taxable_event.crypto_balance_change,
                )
                current_taxable_event_fraction += 1
            else:
                raise RP2ValueError(
                    f"Current taxable event amount ({current_taxable_event_amount})"
                    f" exceeded crypto balance change of taxable event ({gain_loss.taxable_event.crypto_balance_change})"
                    f". {gain_loss}"
                )

            if gain_loss.acquired_lot:
                current_acquired_lot_amount[gain_loss.acquired_lot] = (
                    current_acquired_lot_amount.setdefault(gain_loss.acquired_lot, ZERO) + gain_loss.crypto_amount
                )
                self.__acquired_lots_to_fraction[gain_loss] = current_acquired_lot_fraction.setdefault(gain_loss.acquired_lot, 0)
                if current_acquired_lot_amount[gain_loss.acquired_lot] == gain_loss.acquired_lot.crypto_balance_change:
                    # Expected amount reached: delete both fraction and amount from "current" dictionaries
                    if gain_loss.acquired_lot in self.__acquired_lots_to_number_of_fractions:
                        raise RP2ValueError(f"Acquired lot crypto amount already exhausted for {gain_loss.acquired_lot}")
                    self.__acquired_lots_to_number_of_fractions[gain_loss.acquired_lot] = current_acquired_lot_fraction[gain_loss.acquired_lot] + 1
                    LOGGER.debug(
                        "%s (%d - %d): current amount == acquired lot amount (%.16f)",
                        gain_loss.internal_id,
                        current_acquired_lot_fraction[gain_loss.acquired_lot],
                        current_taxable_event_fraction,
                        current_acquired_lot_amount[gain_loss.acquired_lot],
                    )
                    del current_acquired_lot_amount[gain_loss.acquired_lot]
                    del current_acquired_lot_fraction[gain_loss.acquired_lot]
                elif current_acquired_lot_amount[gain_loss.acquired_lot] < gain_loss.acquired_lot.crypto_balance_change:
                    LOGGER.debug(
                        "%s (%d - %d): current amount < acquired lot amount (%.16f < %.16f)",
                        gain_loss.internal_id,
                        current_acquired_lot_fraction[gain_loss.acquired_lot],
                        current_taxable_event_fraction,
                        current_acquired_lot_amount[gain_loss.acquired_lot],
                        gain_loss.acquired_lot.crypto_balance_change,
                    )
                    current_acquired_lot_fraction[gain_loss.acquired_lot] = current_acquired_lot_fraction[gain_loss.acquired_lot] + 1
                else:
                    raise RP2ValueError(
                        f"Current acquired lot amount ({current_acquired_lot_amount[gain_loss.acquired_lot]}) "
                        f"exceeded crypto balance change of acquired lot ({gain_loss.acquired_lot.crypto_balance_change})"
                        f". {gain_loss}"
                    )

        # Final housekeeping

        # Taxable event: update fractions for last non-exhausted transaction (if any)
        if last_gain_loss_with_acquired_lot:
            if current_taxable_event_amount > ZERO:
                if last_gain_loss_with_acquired_lot.taxable_event in self.__taxable_events_to_number_of_fractions:
                    raise RP2ValueError(f"Taxable event crypto amount already exhausted for {last_gain_loss_with_acquired_lot.taxable_event}")
                self.__taxable_events_to_number_of_fractions[last_gain_loss_with_acquired_lot.taxable_event] = current_taxable_event_fraction
                LOGGER.debug(
                    "%s (%d - %d): taxable event housekeeping",
                    last_gain_loss_with_acquired_lot.internal_id,
                    current_acquired_lot_fraction[last_gain_loss_with_acquired_lot.acquired_lot]
                    if last_gain_loss_with_acquired_lot.acquired_lot in current_acquired_lot_fraction
                    else 0,
                    current_taxable_event_fraction,
                )

        # Acquired lot: update fractions for non-exhausted transactions (if any)
        for acquired_lot, fraction in current_acquired_lot_fraction.items():
            if acquired_lot:
                if acquired_lot in self.__acquired_lots_to_number_of_fractions:
                    raise RP2ValueError(f"Acquired lot crypto amount already exhausted for {acquired_lot}")
                self.__acquired_lots_to_number_of_fractions[acquired_lot] = fraction
                LOGGER.debug(
                    "%s (%d): acquired_lot housekeeping",
                    acquired_lot.internal_id,
                    current_acquired_lot_fraction[acquired_lot],
                )

    def __str__(self) -> str:
        output: List[str] = []
        output.append(f"{type(self).__name__}:")
        output.append(f"  configuration={self.configuration.configuration_path}")
        output.append(f"  asset={self.asset}")
        output.append(f"  from_date={str(self.from_date) if self.from_date > MIN_DATE else 'non-specified'}")
        output.append(f"  to_date={str(self.to_date) if self.to_date < MAX_DATE else 'non-specified'}")
        output.append("  entries=")
        for entry in self:
            parent: Optional[AbstractEntry]
            gain_loss: GainLoss = cast(GainLoss, entry)
            output.append(entry.to_string(indent=2, repr_format=False))
            parent = self.get_parent(entry)
            output.append(
                f"      taxable_event_fraction={self.get_taxable_event_fraction(gain_loss) + 1} of "
                f"{self.get_taxable_event_number_of_fractions(gain_loss.taxable_event)}"
            )
            if gain_loss.acquired_lot:
                output.append(
                    f"      acquired_lot_fraction={self.get_acquired_lot_fraction(gain_loss) + 1} of "
                    f"{self.get_acquired_lot_number_of_fractions(gain_loss.acquired_lot)}"
                )
            output.append(f"      parent={parent.internal_id if parent else None}")
        return "\n".join(output)

    def __repr__(self) -> str:
        output: List[str] = []
        output.append(f"{type(self).__name__}(")
        output.append(f"configuration={repr(self.configuration.configuration_path)}")
        output.append(f", asset={repr(self.asset)}")
        output.append(f", from_date={repr(self.from_date) if self.from_date > MIN_DATE else 'non-specified'}")
        output.append(f", to_date={repr(self.to_date) if self.to_date < MAX_DATE else 'non-specified'}")
        output.append(", entries=[")
        count: int = 0
        for entry in self:
            parent: Optional[AbstractEntry]
            gain_loss: GainLoss = cast(GainLoss, entry)
            if count > 0:
                output.append(", ")
            entry_string = repr(entry)
            # Remove trailing ')' to add set-specific information like parent
            if entry_string[-1] != ")":
                raise RP2RuntimeError("Internal error: repr() of transaction doesn't end with ')'")
            output.append(entry_string[:-1])
            parent = self.get_parent(entry)
            output.append(
                f", taxable_event_fraction={self.get_taxable_event_fraction(gain_loss) + 1} of "
                f"{self.get_taxable_event_number_of_fractions(gain_loss.taxable_event)}"
            )
            if gain_loss.acquired_lot:
                output.append(
                    f", acquired_lot_fraction={self.get_acquired_lot_fraction(gain_loss) + 1} of "
                    f"{self.get_acquired_lot_number_of_fractions(gain_loss.acquired_lot)}"
                )
            output.append(f", parent={parent.internal_id if parent else None}")
            # Add back trailing ')'
            output.append(")")
            count += 1
        output.append("]")
        return "".join(output)



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from rp2.abstract_accounting_method import (
    AbstractAccountingMethod,
    AcquiredLotAndAmount,
    AcquiredLotCandidates,
    AcquiredLotCandidatesOrder,
)
from rp2.abstract_transaction import AbstractTransaction
from rp2.in_transaction import InTransaction
from rp2.rp2_decimal import ZERO, RP2Decimal


# FIFO plugin. See https://www.investopedia.com/terms/l/fifo.asp. This plugin uses universal application, not per-wallet application:
# this means there is one queue for each coin across every wallet and exchange and the accounting method is applied to each such queue.
# More on this at https://www.forbes.com/sites/shehanchandrasekera/2020/09/17/what-crypto-taxpayers-need-to-know-about-fifo-lifo-hifo-specific-id/
class AccountingMethod(AbstractAccountingMethod):
    def seek_non_exhausted_acquired_lot(
        self,
        lot_candidates: AcquiredLotCandidates,
        taxable_event: Optional[AbstractTransaction],
        taxable_event_amount: RP2Decimal,
    ) -> Optional[AcquiredLotAndAmount]:
        selected_acquired_lot_amount: RP2Decimal = ZERO
        selected_acquired_lot: Optional[InTransaction] = None
        acquired_lot: InTransaction
        # This loop causes O(m*n) complexity, where m is the number of acquired lots and n in the number of taxable events. The taxable
        # event loop is in the caller. Non-trivial optimizations are possible using different data structures but they need to be researched.
        for acquired_lot in lot_candidates:
            acquired_lot_amount: RP2Decimal = ZERO

            if not lot_candidates.has_partial_amount(acquired_lot):
                acquired_lot_amount = acquired_lot.crypto_in
            elif lot_candidates.get_partial_amount(acquired_lot) > ZERO:
                acquired_lot_amount = lot_candidates.get_partial_amount(acquired_lot)
            else:
                # The acquired lot has zero partial amount
                continue

            selected_acquired_lot_amount = acquired_lot_amount
            selected_acquired_lot = acquired_lot
            break

        if selected_acquired_lot_amount > ZERO and selected_acquired_lot:
            lot_candidates.clear_partial_amount(selected_acquired_lot)
            return AcquiredLotAndAmount(acquired_lot=selected_acquired_lot, amount=selected_acquired_lot_amount)
        return None

    def lot_candidates_order(self) -> AcquiredLotCandidatesOrder:
        return AcquiredLotCandidatesOrder.OLDER_TO_NEWER



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gettext
import os
from pathlib import Path

from babel import Locale
from babel.core import UnknownLocaleError

from rp2.rp2_error import RP2TypeError, RP2ValueError

_ = gettext.gettext


def set_generation_language(generation_language: str) -> None:
    global _  # pylint: disable=global-statement
    locales_dir = Path(os.path.dirname(__file__)).absolute() / Path("locales")
    if not isinstance(generation_language, str):
        raise RP2TypeError(f"generation_language parameter is not a string: {generation_language}")
    try:
        Locale(generation_language)
        translation = gettext.translation("messages", localedir=str(locales_dir), languages=[generation_language])
    except UnknownLocaleError as exc:
        raise RP2ValueError(f"Unrecognized language: {generation_language}") from exc
    except FileNotFoundError as exc:
        raise RP2ValueError(f"No translation found for language: {generation_language}") from exc

    translation.install()
    _ = translation.gettext



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import unittest
from datetime import date
from pathlib import Path

from abstract_test_ods_output_diff import AbstractTestODSOutputDiff, OutputPlugins

ROOT_PATH: Path = Path(os.path.dirname(__file__)).parent.absolute()


class TestODSOutputDiff(AbstractTestODSOutputDiff):  # pylint: disable=too-many-public-methods
    output_dir: Path

    @classmethod
    def setUpClass(cls) -> None:
        cls.output_dir = ROOT_PATH / Path("output") / Path(cls.__module__)

        shutil.rmtree(cls.output_dir, ignore_errors=True)

        for method in AbstractTestODSOutputDiff.METHODS:
            AbstractTestODSOutputDiff._generate(
                cls.output_dir, test_name="crypto_example", config="crypto_example", method=method, allow_negative_balances=True
            )
            AbstractTestODSOutputDiff._generate(cls.output_dir, test_name="test_data", config="test_data", method=method)
            AbstractTestODSOutputDiff._generate(cls.output_dir, test_name="test_data2", config="test_data", method=method, allow_negative_balances=True)
            AbstractTestODSOutputDiff._generate(cls.output_dir, test_name="test_data3", config="test_data", method=method)
            AbstractTestODSOutputDiff._generate(cls.output_dir, test_name="test_data4", config="test_data4", method=method)
            AbstractTestODSOutputDiff._generate(cls.output_dir, test_name="test_hifo", config="test_data", method=method, allow_negative_balances=True)
            AbstractTestODSOutputDiff._generate(cls.output_dir, test_name="test_hifo2", config="test_data", method=method, allow_negative_balances=True)
            AbstractTestODSOutputDiff._generate(
                cls.output_dir, test_name="test_many_year_data", config="test_data", method=method, allow_negative_balances=True
            )
            AbstractTestODSOutputDiff._generate(
                cls.output_dir,
                test_name="test_data3",
                config="test_data",
                method=method,
                from_date=date(2019, 12, 1),
                to_date=date(2020, 4, 1),
                allow_negative_balances=True,
            )

        AbstractTestODSOutputDiff._generate(
            cls.output_dir, test_name="test_many_year_data", config="test_data", method="fifo", to_date=date(2016, 12, 31), allow_negative_balances=True
        )
        AbstractTestODSOutputDiff._generate(
            cls.output_dir, test_name="test_many_year_data", config="test_data", method="fifo", to_date=date(2017, 12, 31), allow_negative_balances=True
        )
        AbstractTestODSOutputDiff._generate(
            cls.output_dir, test_name="test_many_year_data", config="test_data", method="fifo", to_date=date(2018, 12, 31), allow_negative_balances=True
        )
        AbstractTestODSOutputDiff._generate(
            cls.output_dir, test_name="test_many_year_data", config="test_data", method="fifo", to_date=date(2019, 12, 31), allow_negative_balances=True
        )
        AbstractTestODSOutputDiff._generate(
            cls.output_dir, test_name="test_many_year_data", config="test_data", method="fifo", to_date=date(2020, 12, 31), allow_negative_balances=True
        )

        AbstractTestODSOutputDiff._generate(
            cls.output_dir, test_name="test_many_year_data", config="test_data", method="fifo", from_date=date(2017, 1, 1), allow_negative_balances=True
        )
        AbstractTestODSOutputDiff._generate(
            cls.output_dir, test_name="test_many_year_data", config="test_data", method="fifo", from_date=date(2018, 1, 1), allow_negative_balances=True
        )
        AbstractTestODSOutputDiff._generate(
            cls.output_dir, test_name="test_many_year_data", config="test_data", method="fifo", from_date=date(2019, 1, 1), allow_negative_balances=True
        )
        AbstractTestODSOutputDiff._generate(
            cls.output_dir, test_name="test_many_year_data", config="test_data", method="fifo", from_date=date(2020, 1, 1), allow_negative_balances=True
        )
        AbstractTestODSOutputDiff._generate(
            cls.output_dir, test_name="test_many_year_data", config="test_data", method="fifo", from_date=date(2021, 1, 1), allow_negative_balances=True
        )

        AbstractTestODSOutputDiff._generate(
            cls.output_dir,
            test_name="test_many_year_data",
            config="test_data",
            method="fifo",
            from_date=date(2017, 1, 1),
            to_date=date(2019, 12, 31),
            allow_negative_balances=True,
        )
        AbstractTestODSOutputDiff._generate(
            cls.output_dir,
            test_name="test_many_year_data",
            config="test_data",
            method="fifo",
            from_date=date(2018, 1, 1),
            to_date=date(2019, 12, 31),
            allow_negative_balances=True,
        )
        AbstractTestODSOutputDiff._generate(
            cls.output_dir,
            test_name="test_many_year_data",
            config="test_data",
            method="fifo",
            from_date=date(2019, 1, 1),
            to_date=date(2019, 12, 31),
            allow_negative_balances=True,
        )
        AbstractTestODSOutputDiff._generate(
            cls.output_dir, test_name="test_data_multi_method", config="test_data_multi_method", method="mixed", allow_negative_balances=True
        )

    def setUp(self) -> None:
        self.maxDiff = None  # pylint: disable=invalid-name

    def test_crypto_example_rp2_full_report(self) -> None:
        for method in self.METHODS:
            self._compare(output_dir=self.output_dir, test_name="crypto_example", method=method, output_plugin=OutputPlugins.RP2_FULL_REPORT)

    def test_crypto_example_tax_report_us(self) -> None:
        for method in self.METHODS:
            self._compare(output_dir=self.output_dir, test_name="crypto_example", method=method, output_plugin=OutputPlugins.TAX_REPORT_US)

    def test_test_data_rp2_full_report(self) -> None:
        for method in self.METHODS:
            self._compare(output_dir=self.output_dir, test_name="test_data", method=method, output_plugin=OutputPlugins.RP2_FULL_REPORT)

    def test_test_data_tax_report_us(self) -> None:
        for method in self.METHODS:
            self._compare(output_dir=self.output_dir, test_name="test_data", method=method, output_plugin=OutputPlugins.TAX_REPORT_US)

    def test_test_data2_rp2_full_report(self) -> None:
        for method in self.METHODS:
            self._compare(output_dir=self.output_dir, test_name="test_data2", method=method, output_plugin=OutputPlugins.RP2_FULL_REPORT)

    def test_test_data2_tax_report_us(self) -> None:
        for method in self.METHODS:
            self._compare(output_dir=self.output_dir, test_name="test_data2", method=method, output_plugin=OutputPlugins.TAX_REPORT_US)

    def test_test_data3_rp2_full_report(self) -> None:
        for method in self.METHODS:
            self._compare(output_dir=self.output_dir, test_name="test_data3", method=method, output_plugin=OutputPlugins.RP2_FULL_REPORT)

    def test_test_data3_tax_report_us(self) -> None:
        for method in self.METHODS:
            self._compare(output_dir=self.output_dir, test_name="test_data3", method=method, output_plugin=OutputPlugins.TAX_REPORT_US)

    def test_test_data4_rp2_full_report(self) -> None:
        for method in self.METHODS:
            self._compare(output_dir=self.output_dir, test_name="test_data4", method=method, output_plugin=OutputPlugins.RP2_FULL_REPORT)

    def test_test_data4_tax_report_us(self) -> None:
        for method in self.METHODS:
            self._compare(output_dir=self.output_dir, test_name="test_data4", method=method, output_plugin=OutputPlugins.TAX_REPORT_US)

    def test_test_hifo_rp2_full_report(self) -> None:
        for method in self.METHODS:
            self._compare(output_dir=self.output_dir, test_name="test_hifo", method=method, output_plugin=OutputPlugins.RP2_FULL_REPORT)

    def test_test_hifo_tax_report_us(self) -> None:
        for method in self.METHODS:
            self._compare(output_dir=self.output_dir, test_name="test_hifo", method=method, output_plugin=OutputPlugins.TAX_REPORT_US)

    def test_test_hifo2_rp2_full_report(self) -> None:
        for method in self.METHODS:
            self._compare(output_dir=self.output_dir, test_name="test_hifo2", method=method, output_plugin=OutputPlugins.RP2_FULL_REPORT)

    def test_test_hifo2_tax_report_us(self) -> None:
        for method in self.METHODS:
            self._compare(output_dir=self.output_dir, test_name="test_hifo2", method=method, output_plugin=OutputPlugins.TAX_REPORT_US)

    def test_test_many_year_data_rp2_full_report(self) -> None:
        for method in self.METHODS:
            self._compare(output_dir=self.output_dir, test_name="test_many_year_data", method=method, output_plugin=OutputPlugins.RP2_FULL_REPORT)

    def test_test_many_year_data_tax_report_us(self) -> None:
        for method in self.METHODS:
            self._compare(output_dir=self.output_dir, test_name="test_many_year_data", method=method, output_plugin=OutputPlugins.TAX_REPORT_US)

    def test_test_data3_rp2_full_report_2019_12_01_2020_04_01(self) -> None:
        for method in self.METHODS:
            self._compare(
                output_dir=self.output_dir,
                test_name="test_data3",
                method=method,
                output_plugin=OutputPlugins.RP2_FULL_REPORT,
                from_date=date(2019, 12, 1),
                to_date=date(2020, 4, 1),
            )

    def test_test_data3_tax_report_us_2019_12_01_2020_04_01(self) -> None:
        for method in self.METHODS:
            self._compare(
                output_dir=self.output_dir,
                test_name="test_data3",
                method=method,
                output_plugin=OutputPlugins.TAX_REPORT_US,
                from_date=date(2019, 12, 1),
                to_date=date(2020, 4, 1),
            )

    def test_test_many_year_data_rp2_full_report_0_2016(self) -> None:
        self._compare(
            output_dir=self.output_dir, test_name="test_many_year_data", method="fifo", output_plugin=OutputPlugins.RP2_FULL_REPORT, to_date=date(2016, 12, 31)
        )

    def test_test_many_year_data_tax_report_us_0_2016(self) -> None:
        self._compare(
            output_dir=self.output_dir, test_name="test_many_year_data", method="fifo", output_plugin=OutputPlugins.TAX_REPORT_US, to_date=date(2016, 12, 31)
        )

    def test_test_many_year_data_rp2_full_report_0_2017(self) -> None:
        self._compare(
            output_dir=self.output_dir, test_name="test_many_year_data", method="fifo", output_plugin=OutputPlugins.RP2_FULL_REPORT, to_date=date(2017, 12, 31)
        )

    def test_test_many_year_data_tax_report_us_0_2017(self) -> None:
        self._compare(
            output_dir=self.output_dir, test_name="test_many_year_data", method="fifo", output_plugin=OutputPlugins.TAX_REPORT_US, to_date=date(2017, 12, 31)
        )

    def test_test_many_year_data_rp2_full_report_0_2018(self) -> None:
        self._compare(
            output_dir=self.output_dir, test_name="test_many_year_data", method="fifo", output_plugin=OutputPlugins.RP2_FULL_REPORT, to_date=date(2018, 12, 31)
        )

    def test_test_many_year_data_tax_report_us_0_2018(self) -> None:
        self._compare(
            output_dir=self.output_dir, test_name="test_many_year_data", method="fifo", output_plugin=OutputPlugins.TAX_REPORT_US, to_date=date(2018, 12, 31)
        )

    def test_test_many_year_data_rp2_full_report_0_2019(self) -> None:
        self._compare(
            output_dir=self.output_dir, test_name="test_many_year_data", method="fifo", output_plugin=OutputPlugins.RP2_FULL_REPORT, to_date=date(2019, 12, 31)
        )

    def test_test_many_year_data_tax_report_us_0_2019(self) -> None:
        self._compare(
            output_dir=self.output_dir, test_name="test_many_year_data", method="fifo", output_plugin=OutputPlugins.TAX_REPORT_US, to_date=date(2019, 12, 31)
        )

    def test_test_many_year_data_rp2_full_report_0_2020(self) -> None:
        self._compare(
            output_dir=self.output_dir, test_name="test_many_year_data", method="fifo", output_plugin=OutputPlugins.RP2_FULL_REPORT, to_date=date(2020, 12, 31)
        )

    def test_test_many_year_data_tax_report_us_0_2020(self) -> None:
        self._compare(
            output_dir=self.output_dir, test_name="test_many_year_data", method="fifo", output_plugin=OutputPlugins.TAX_REPORT_US, to_date=date(2020, 12, 31)
        )

    def test_test_many_year_data_rp2_full_report_2017_infinity(self) -> None:
        self._compare(
            output_dir=self.output_dir, test_name="test_many_year_data", method="fifo", output_plugin=OutputPlugins.RP2_FULL_REPORT, from_date=date(2017, 1, 1)
        )

    def test_test_many_year_data_tax_report_us_2017_infinity(self) -> None:
        self._compare(
            output_dir=self.output_dir, test_name="test_many_year_data", method="fifo", output_plugin=OutputPlugins.TAX_REPORT_US, from_date=date(2017, 1, 1)
        )

    def test_test_many_year_data_rp2_full_report_2018_infinity(self) -> None:
        self._compare(
            output_dir=self.output_dir, test_name="test_many_year_data", method="fifo", output_plugin=OutputPlugins.RP2_FULL_REPORT, from_date=date(2018, 1, 1)
        )

    def test_test_many_year_data_tax_report_us_2018_infinity(self) -> None:
        self._compare(
            output_dir=self.output_dir, test_name="test_many_year_data", method="fifo", output_plugin=OutputPlugins.TAX_REPORT_US, from_date=date(2018, 1, 1)
        )

    def test_test_many_year_data_rp2_full_report_2019_infinity(self) -> None:
        self._compare(
            output_dir=self.output_dir, test_name="test_many_year_data", method="fifo", output_plugin=OutputPlugins.RP2_FULL_REPORT, from_date=date(2019, 1, 1)
        )

    def test_test_many_year_data_tax_report_us_2019_infinity(self) -> None:
        self._compare(
            output_dir=self.output_dir, test_name="test_many_year_data", method="fifo", output_plugin=OutputPlugins.TAX_REPORT_US, from_date=date(2019, 1, 1)
        )

    def test_test_many_year_data_rp2_full_report_2020_infinity(self) -> None:
        self._compare(
            output_dir=self.output_dir, test_name="test_many_year_data", method="fifo", output_plugin=OutputPlugins.RP2_FULL_REPORT, from_date=date(2020, 1, 1)
        )

    def test_test_many_year_data_tax_report_us_2020_infinity(self) -> None:
        self._compare(
            output_dir=self.output_dir, test_name="test_many_year_data", method="fifo", output_plugin=OutputPlugins.TAX_REPORT_US, from_date=date(2020, 1, 1)
        )

    def test_test_many_year_data_rp2_full_report_2021_infinity(self) -> None:
        self._compare(
            output_dir=self.output_dir, test_name="test_many_year_data", method="fifo", output_plugin=OutputPlugins.RP2_FULL_REPORT, from_date=date(2021, 1, 1)
        )

    def test_test_many_year_data_tax_report_us_2021_infinity(self) -> None:
        self._compare(
            output_dir=self.output_dir, test_name="test_many_year_data", method="fifo", output_plugin=OutputPlugins.TAX_REPORT_US, from_date=date(2021, 1, 1)
        )

    def test_test_many_year_data_rp2_full_report_2017_2019(self) -> None:
        self._compare(
            output_dir=self.output_dir,
            test_name="test_many_year_data",
            method="fifo",
            output_plugin=OutputPlugins.RP2_FULL_REPORT,
            from_date=date(2017, 1, 1),
            to_date=date(2019, 12, 31),
        )

    def test_test_many_year_data_tax_report_us_2017_2019(self) -> None:
        self._compare(
            output_dir=self.output_dir,
            test_name="test_many_year_data",
            method="fifo",
            output_plugin=OutputPlugins.TAX_REPORT_US,
            from_date=date(2017, 1, 1),
            to_date=date(2019, 12, 31),
        )

    def test_test_many_year_data_rp2_full_report_2018_2019(self) -> None:
        self._compare(
            output_dir=self.output_dir,
            test_name="test_many_year_data",
            method="fifo",
            output_plugin=OutputPlugins.RP2_FULL_REPORT,
            from_date=date(2018, 1, 1),
            to_date=date(2019, 12, 31),
        )

    def test_test_many_year_data_tax_report_us_2018_2019(self) -> None:
        self._compare(
            output_dir=self.output_dir,
            test_name="test_many_year_data",
            method="fifo",
            output_plugin=OutputPlugins.TAX_REPORT_US,
            from_date=date(2018, 1, 1),
            to_date=date(2019, 12, 31),
        )

    def test_test_many_year_data_rp2_full_report_2019_2019(self) -> None:
        self._compare(
            output_dir=self.output_dir,
            test_name="test_many_year_data",
            method="fifo",
            output_plugin=OutputPlugins.RP2_FULL_REPORT,
            from_date=date(2019, 1, 1),
            to_date=date(2019, 12, 31),
        )

    def test_test_many_year_data_tax_report_us_2019_2019(self) -> None:
        self._compare(
            output_dir=self.output_dir,
            test_name="test_many_year_data",
            method="fifo",
            output_plugin=OutputPlugins.TAX_REPORT_US,
            from_date=date(2019, 1, 1),
            to_date=date(2019, 12, 31),
        )

    def test_test_data_multi_method_rp2_full_report(self) -> None:
        self._compare(
            output_dir=self.output_dir,
            test_name="test_data_multi_method",
            method="mixed",
            output_plugin=OutputPlugins.RP2_FULL_REPORT,
        )

    def test_test_data_multi_method_tax_report_us(self) -> None:
        self._compare(
            output_dir=self.output_dir,
            test_name="test_data_multi_method",
            method="mixed",
            output_plugin=OutputPlugins.TAX_REPORT_US,
        )


if __name__ == "__main__":
    unittest.main()



# Copyright 2022 macanudo527
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
from typing import Set

from rp2.abstract_country import AbstractCountry
from rp2.rp2_main import rp2_main


# JP-specific class
class JP(AbstractCountry):
    def __init__(self) -> None:
        super().__init__("jp", "jpy")

    # Measured in days
    def get_long_term_capital_gain_period(self) -> int:
        # No long-term capital gains in Japan for crypto assets (as of 7/2022)
        return sys.maxsize

    # Default accounting method to use if the user doesn't specify one on the command line
    def get_default_accounting_method(self) -> str:
        # This is incorrect and only a placeholder: we still need to implement Japan-specific accounting methods
        return "fifo"

    # Set of accounting methods accepted in the country
    def get_accounting_methods(self) -> Set[str]:
        # This is incorrect and only a placeholder: we still need to implement Japan-specific accounting methods
        return {"fifo"}

    def get_report_generators(self) -> Set[str]:
        return {
            "open_positions",
            "rp2_full_report",
            "jp.tax_report_jp",
        }

    # Default language to use at report generation if the user doesn't specify it on the command line (in ISO 639-1 format)
    def get_default_generation_language(self) -> str:
        return "ja"


# JP-specific entry point
def rp2_entry() -> None:
    rp2_main(JP())



# Copyright 2022 Neal Chambers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from datetime import date
from enum import Enum
from itertools import chain
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, Set, cast

from rp2.abstract_country import AbstractCountry
from rp2.abstract_entry import AbstractEntry
from rp2.abstract_transaction import AbstractTransaction
from rp2.computed_data import ComputedData
from rp2.configuration import MAX_DATE, MIN_DATE
from rp2.entry_types import TransactionType
from rp2.in_transaction import InTransaction
from rp2.intra_transaction import IntraTransaction
from rp2.localization import _
from rp2.logger import create_logger
from rp2.out_transaction import OutTransaction
from rp2.plugin.report.abstract_ods_generator import AbstractODSGenerator
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2RuntimeError, RP2TypeError
from rp2.transaction_set import TransactionSet

LOGGER: logging.Logger = create_logger("tax_report_jp")


class _TransactionRow(NamedTuple):
    transaction_type: str
    transaction_month: int
    transaction_day: int
    transaction_client: str
    fee_in_yen: RP2Decimal = ZERO
    gift: RP2Decimal = ZERO
    purchase_crypto_amount: Optional[RP2Decimal] = None
    purchase_amount_in_yen: Optional[RP2Decimal] = None
    sales_crypto_amount: Optional[RP2Decimal] = None
    sales_amount_in_yen: Optional[RP2Decimal] = None
    donated_amount_in_yen: Optional[RP2Decimal] = None


class _SheetNames(Enum):
    ASSET: str = "Asset"
    SUMMARY: str = "Summary"


_TEMPLATE_SHEETS_TO_KEEP: Set[str] = {f"__{item.value}" for item in _SheetNames}
_INCOME_TRANSACTION_TYPES: Dict[TransactionType, None] = {
    TransactionType.AIRDROP: None,
    TransactionType.HARDFORK: None,
    TransactionType.INCOME: None,
    TransactionType.INTEREST: None,
    TransactionType.MINING: None,
    TransactionType.STAKING: None,
    TransactionType.WAGES: None,
}


class Generator(AbstractODSGenerator):
    MIN_ROWS: int = 20
    MAX_COLUMNS: int = 20
    OUTPUT_FILE: str = "tax_report_jp.ods"

    ASSET_TEMPLATE_SHEET: str = "Asset"
    SUMMARY_TEMPLATE_SHEET: str = "Summary"
    TRANSACTION_ROW_START: int = 22

    TRANSFER: str = "Transfer"

    def __init__(self) -> None:
        super().__init__()
        self.__year_row_offset: Dict[int, int] = {}
        self.__number_of_summaries: int = 0

    def generate(
        self,
        country: AbstractCountry,
        years_2_accounting_method_names: Dict[int, str],
        asset_to_computed_data: Dict[str, ComputedData],
        output_dir_path: str,
        output_file_prefix: str,
        from_date: date,
        to_date: date,
        generation_language: str,
    ) -> None:
        if from_date != MIN_DATE and to_date != MAX_DATE:
            raise RP2RuntimeError("To and From Dates can not be specified for the JP tax report.")

        if not isinstance(asset_to_computed_data, Dict):
            raise RP2TypeError(f"Parameter 'asset_to_computed_data' has non-Dict value {asset_to_computed_data}")

        template_path: str = self._get_template_path("tax_report_jp", country, generation_language)

        output_file: Any
        output_file = self._initialize_output_file(
            country=country,
            legend_data=[],
            years_2_accounting_method_names=years_2_accounting_method_names,
            output_dir_path=output_dir_path,
            output_file_prefix=output_file_prefix,
            output_file_name=self.OUTPUT_FILE,
            template_path=template_path,
            template_sheets_to_keep=_TEMPLATE_SHEETS_TO_KEEP,
            from_date=from_date,
            to_date=to_date,
        )

        asset: str
        computed_data: ComputedData

        for asset, computed_data in asset_to_computed_data.items():
            if not isinstance(asset, str):
                raise RP2TypeError(f"Parameter 'asset' has non-string value {asset}")
            ComputedData.type_check("computed_data", computed_data)
            self.__generate_asset(computed_data, output_file)

        del output_file.sheets[self.ASSET_TEMPLATE_SHEET]
        del output_file.sheets[self.SUMMARY_TEMPLATE_SHEET]

        output_file.save()
        LOGGER.info("Plugin '%s' output: %s", __name__, Path(output_file.docname).resolve())

    def __insert_secondary_transaction_row(self, sheet_name: Any, row: int) -> None:
        sheet_name.insert_rows(index=row, count=1)
        sheet_name[f"A{row + 1}"].style_name = "transactions_month"
        sheet_name[f"B{row + 1}"].style_name = "transactions_day"
        sheet_name[f"C{row + 1}"].style_name = "transactions_middle"
        sheet_name[f"D{row + 1}"].style_name = "transactions_middle"
        sheet_name[f"E{row + 1}"].style_name = "transactions_crypto"
        sheet_name[f"F{row + 1}"].style_name = "transactions_yen"
        sheet_name[f"G{row + 1}"].style_name = "transactions_crypto"
        sheet_name[f"H{row + 1}"].style_name = "transactions_yen"
        sheet_name[f"I{row + 1}"].style_name = "transactions_right_end"

    def __insert_summary_row(self, sheet_name: Any, row: int) -> None:
        sheet_name.insert_rows(index=row, count=1)
        sheet_name[f"A{row + 1}"].style_name = "transactions_left_end"
        sheet_name[f"B{row + 1}"].style_name = "transactions_middle"
        sheet_name[f"C{row + 1}"].style_name = "transactions_middle"
        sheet_name[f"D{row + 1}"].style_name = "transactions_middle"
        sheet_name[f"E{row + 1}"].style_name = "transactions_crypto"
        sheet_name[f"F{row + 1}"].style_name = "transactions_yen"
        sheet_name[f"G{row + 1}"].style_name = "transactions_right_end"

    @staticmethod
    def get_tax_sheet_name(asset: str, year: int) -> str:
        return _("{}_{}").format(asset, year)

    @staticmethod
    def get_summary_sheet_name(year: int) -> str:
        return _("{}_Summary").format(year)

    def __generate_asset(self, computed_data: ComputedData, output_file: Any) -> None:
        asset: str = computed_data.asset

        in_transaction_set: TransactionSet = computed_data.in_transaction_set
        out_transaction_set: TransactionSet = computed_data.out_transaction_set
        intra_transaction_set: TransactionSet = computed_data.intra_transaction_set
        entry: AbstractEntry
        year: int
        years_2_transaction_sets: Dict[int, List[AbstractTransaction]] = {}
        previous_year_row_offset: int = 0

        # Sort all in and out transactions by year, the fee from intra transactions must be reported
        for entry in chain(in_transaction_set, out_transaction_set, intra_transaction_set):  # type: ignore
            transaction: AbstractTransaction = cast(AbstractTransaction, entry)
            years_2_transaction_sets.setdefault(transaction.timestamp.year, []).append(entry)

        for year, transaction_set in years_2_transaction_sets.items():
            # Sort the transactions by timestamp and generate sheet by year
            previous_year_row_offset = self.__generate_asset_year(
                asset=asset,
                year=year,
                transaction_list=sorted(transaction_set, key=lambda x: x.timestamp),
                output_file=output_file,
                previous_year_row_offset=previous_year_row_offset,
            )

            summary_sheet: Any = output_file.sheets[self.get_summary_sheet_name(year)]

            ### Totals at the bottom
            # Amount Donated
            self._fill_cell(summary_sheet, self.__year_row_offset[year] + 2, 1, f"=SUM(B8:B{self.__year_row_offset[year] + 2})", apply_style=False)
            # Gift Received
            self._fill_cell(summary_sheet, self.__year_row_offset[year] + 2, 2, f"=SUM(C8:C{self.__year_row_offset[year] + 2})", apply_style=False)
            # Total Amount in yen
            self._fill_cell(summary_sheet, self.__year_row_offset[year] + 2, 5, f"=SUM(F8:F{self.__year_row_offset[year] + 2})", apply_style=False)
            # Net Income Amt
            self._fill_cell(summary_sheet, self.__year_row_offset[year] + 2, 6, f"=SUM(G8:G{self.__year_row_offset[year] + 2})", apply_style=False)

    def __process_in_transaction(self, transaction: InTransaction) -> _TransactionRow:
        purchase_amount_in_yen: RP2Decimal = transaction.crypto_in * transaction.spot_price

        # Find the fee in yen
        fee_in_yen: RP2Decimal = ZERO
        if RP2Decimal(transaction.crypto_fee) > ZERO:
            fee_in_yen = transaction.crypto_fee * transaction.spot_price
        elif RP2Decimal(transaction.fiat_fee) > ZERO:
            fee_in_yen = transaction.fiat_fee

        # Record the profit this year of crypto assets earned as income
        # DONATE and GIFT (up to what is allowed) is tax free so we don't report them as income
        sales_crypto_amount: Optional[RP2Decimal] = None
        sales_amount_in_yen: Optional[RP2Decimal] = None
        if transaction.transaction_type in _INCOME_TRANSACTION_TYPES:
            sales_crypto_amount = ZERO
            sales_amount_in_yen = purchase_amount_in_yen

        gift: RP2Decimal = ZERO
        if transaction.transaction_type == TransactionType.GIFT:
            gift = purchase_amount_in_yen

        return _TransactionRow(
            transaction_type=transaction.transaction_type.value.upper(),
            purchase_crypto_amount=transaction.crypto_in,
            purchase_amount_in_yen=purchase_amount_in_yen,
            transaction_month=transaction.timestamp.month,
            transaction_day=transaction.timestamp.day,
            transaction_client=transaction.exchange,
            sales_crypto_amount=sales_crypto_amount,
            sales_amount_in_yen=sales_amount_in_yen,
            fee_in_yen=fee_in_yen,
            gift=gift,
        )

    def __process_intra_transaction(self, transaction: IntraTransaction) -> _TransactionRow:
        transaction_fee_in_crypto: Optional[RP2Decimal] = None
        transaction_fee_in_yen: Optional[RP2Decimal] = None

        transaction_fee_in_crypto = transaction.crypto_sent - transaction.crypto_received
        transaction_fee_in_yen = transaction_fee_in_crypto * transaction.spot_price

        return _TransactionRow(
            transaction_type=TransactionType.FEE.value.upper(),
            transaction_month=transaction.timestamp.month,
            transaction_day=transaction.timestamp.day,
            transaction_client=_(self.TRANSFER),
            sales_crypto_amount=transaction_fee_in_crypto if transaction_fee_in_crypto > ZERO else None,
            sales_amount_in_yen=transaction_fee_in_yen if transaction_fee_in_yen > ZERO else None,
            fee_in_yen=ZERO,
            gift=ZERO,
        )

    def __process_out_transaction(self, transaction: OutTransaction) -> _TransactionRow:
        # Find the fee in yen
        fee_in_yen: RP2Decimal = ZERO
        if RP2Decimal(transaction.crypto_fee) > ZERO:
            fee_in_yen = transaction.crypto_fee * transaction.spot_price
        elif RP2Decimal(transaction.fiat_fee) > ZERO:
            fee_in_yen = transaction.fiat_fee

        # DONATE can be used to reduce tax burden, GIFT is taxed as if sold at the moment it is gifted
        donated_amount_in_yen: Optional[RP2Decimal] = None
        sales_amount_in_yen: Optional[RP2Decimal] = None
        if transaction.transaction_type == TransactionType.DONATE:
            donated_amount_in_yen = RP2Decimal(str(transaction.crypto_out_no_fee * transaction.spot_price))
            sales_amount_in_yen = RP2Decimal("0")
        else:
            sales_amount_in_yen = transaction.crypto_out_no_fee * transaction.spot_price

        return _TransactionRow(
            transaction_type=transaction.transaction_type.value.upper(),
            transaction_month=transaction.timestamp.month,
            transaction_day=transaction.timestamp.day,
            transaction_client=transaction.exchange,
            sales_crypto_amount=transaction.crypto_out_with_fee,
            sales_amount_in_yen=sales_amount_in_yen,
            fee_in_yen=fee_in_yen,
            donated_amount_in_yen=donated_amount_in_yen,
        )

    def __generate_asset_year(self, asset: str, year: int, transaction_list: List[AbstractTransaction], output_file: Any, previous_year_row_offset: int) -> int:
        asset_year_sheet: Any = output_file.sheets[self.ASSET_TEMPLATE_SHEET].copy(newname=self.get_tax_sheet_name(asset, year))
        output_file.sheets += asset_year_sheet

        # Label the asset at the top of the page
        self._fill_cell(asset_year_sheet, 1, 7, asset, apply_style=False)
        row_index: int = 21
        total_donations: RP2Decimal = ZERO
        total_gifts: RP2Decimal = ZERO
        formatted_donation_amount: Optional[str] = None
        transaction_row: _TransactionRow

        for entry in transaction_list:
            if isinstance(entry, InTransaction):
                transaction_row = self.__process_in_transaction(entry)
                total_gifts += transaction_row.gift

            elif isinstance(entry, OutTransaction):
                transaction_row = self.__process_out_transaction(entry)
                if transaction_row.donated_amount_in_yen is not None:
                    total_donations += transaction_row.donated_amount_in_yen
                    formatted_donation_amount = f"0 (￥{float(transaction_row.donated_amount_in_yen):0,.2f})"

            elif isinstance(entry, IntraTransaction):
                transaction_row = self.__process_intra_transaction(entry)

            else:
                raise RP2RuntimeError("Transaction is not InTransaction or OutTransaction.")

            if transaction_row.purchase_crypto_amount is None and transaction_row.sales_crypto_amount is None:
                continue

            self.__insert_secondary_transaction_row(asset_year_sheet, row_index)
            self._fill_cell(asset_year_sheet, row_index, 0, transaction_row.transaction_month, apply_style=False)
            self._fill_cell(asset_year_sheet, row_index, 1, transaction_row.transaction_day, apply_style=False)
            self._fill_cell(asset_year_sheet, row_index, 2, transaction_row.transaction_client, apply_style=False)
            self._fill_cell(asset_year_sheet, row_index, 3, transaction_row.transaction_type, apply_style=False)
            if transaction_row.purchase_crypto_amount is not None:
                self._fill_cell(asset_year_sheet, row_index, 4, transaction_row.purchase_crypto_amount, apply_style=False)
                self._fill_cell(asset_year_sheet, row_index, 5, transaction_row.purchase_amount_in_yen, apply_style=False)
            if transaction_row.sales_crypto_amount is not None:
                self._fill_cell(asset_year_sheet, row_index, 6, transaction_row.sales_crypto_amount, apply_style=False)
                self._fill_cell(
                    asset_year_sheet,
                    row_index,
                    7,
                    transaction_row.sales_amount_in_yen if formatted_donation_amount is None else formatted_donation_amount,
                    apply_style=False,
                )
            self._fill_cell(asset_year_sheet, row_index, 8, transaction_row.fee_in_yen, apply_style=False)

            row_index += 1
            formatted_donation_amount = None

        # Adding the summary formulas at the bottom
        self._fill_cell(asset_year_sheet, row_index + 2, 4, f"=IF(SUM(E22:E{row_index+2})=0;0;SUM(E22:E{row_index+2}))", apply_style=False)
        self._fill_cell(asset_year_sheet, row_index + 2, 5, f"=IF(SUM(F22:F{row_index+2})=0;0;SUM(F22:F{row_index+2}))", apply_style=False)
        self._fill_cell(asset_year_sheet, row_index + 2, 6, f"=IF(SUM(G22:G{row_index+2})=0;0;SUM(G22:G{row_index+2}))", apply_style=False)
        self._fill_cell(asset_year_sheet, row_index + 2, 7, f"=IF(SUM(H22:H{row_index+2})=0;0;SUM(H22:H{row_index+2}))", apply_style=False)
        self._fill_cell(asset_year_sheet, row_index + 2, 8, f"=IF(SUM(I22:I{row_index+2})=0;0;SUM(I22:I{row_index+2}))", apply_style=False)

        ### 4 Calculation of Cost of Sale of Cryptographic Assets (暗号資産の売却原価の計算)

        # Last year's totals
        previous_year_crypto_cell: Optional[str] = None
        previous_year_yen_cell: Optional[str] = None
        if previous_year_row_offset != 0:
            previous_year_sheet_name: str = self.get_tax_sheet_name(asset, year - 1)
            previous_year_crypto_cell = f"='{previous_year_sheet_name}'.I{previous_year_row_offset}"
            previous_year_yen_cell = f"='{previous_year_sheet_name}'.I{previous_year_row_offset+1}"

        self._fill_cell(
            asset_year_sheet,
            row_index + 8,
            4,
            previous_year_crypto_cell if previous_year_crypto_cell else 0,
            apply_style=False,
        )
        self._fill_cell(
            asset_year_sheet,
            row_index + 9,
            4,
            previous_year_yen_cell if previous_year_yen_cell else 0,
            apply_style=False,
        )

        # This year's total purchases, etc...
        self._fill_cell(asset_year_sheet, row_index + 8, 5, f"=E13+E{row_index+3}", apply_style=False)
        self._fill_cell(asset_year_sheet, row_index + 9, 5, f"=F13+F{row_index+3}", apply_style=False)

        # Average Unit Price, located at G{row_index+10}
        self._fill_cell(
            asset_year_sheet,
            row_index + 9,
            6,
            f"=IF((E{row_index+9}+F{row_index+9});(E{row_index+10}+F{row_index+10})/(E{row_index+9}+F{row_index+9});0)",
            apply_style=False,
        )

        ## This year's total sales, etc...
        # Total units sold
        self._fill_cell(asset_year_sheet, row_index + 8, 7, f"=G13+G{row_index+3}", apply_style=False)
        # Total in yen (total units * average unit price)
        self._fill_cell(asset_year_sheet, row_index + 9, 7, f"=H{row_index+9}*G{row_index+10}", apply_style=False)

        ## This year's totals
        # Total units left
        self._fill_cell(asset_year_sheet, row_index + 8, 8, f"=E{row_index+9}+F{row_index+9}-H{row_index+9}", apply_style=False)
        # Total in yen (total units left * average unit price)
        self._fill_cell(asset_year_sheet, row_index + 9, 8, f"=I{row_index+9}*G{row_index+10}", apply_style=False)

        ### 5 Calculation of Income Amount for Virtual Currency (仮想通貨の所得金額の計算)

        ## Gross Income Amount
        # Sale Price
        self._fill_cell(asset_year_sheet, row_index + 17, 0, f"=H13+H{row_index+3}", apply_style=False)
        # Cost of Sale
        self._fill_cell(asset_year_sheet, row_index + 17, 5, f"=H{row_index+10}", apply_style=False)
        # Total Fees
        self._fill_cell(asset_year_sheet, row_index + 17, 6, f"=I{row_index+3}", apply_style=False)
        # Net Income Amt
        self._fill_cell(asset_year_sheet, row_index + 17, 8, f"=I{row_index+20}-I{row_index+21}", apply_style=False)

        # Total Profit
        self._fill_cell(asset_year_sheet, row_index + 19, 8, f"=ROUNDDOWN(SUM(A{row_index+18}:E{row_index+18});0)", apply_style=False)
        # Total Necessary Expenses
        self._fill_cell(asset_year_sheet, row_index + 20, 8, f"=ROUNDUP(SUM(F{row_index+18}:H{row_index+18});0)", apply_style=False)

        ### Add to Summary
        year_summary_sheet: Any
        # If this is the first time we have checked the year offset, set it to 7 and create a new summary sheet
        if self.__year_row_offset.setdefault(year, 7) == 7:
            year_summary_sheet = output_file.sheets[self.SUMMARY_TEMPLATE_SHEET].copy(newname=self.get_summary_sheet_name(year))
            # Place summaries at the beginning so they are easy to find
            output_file.sheets.insert(2 + self.__number_of_summaries, year_summary_sheet)
            self.__number_of_summaries += 1
        else:
            year_summary_sheet = output_file.sheets[self.get_summary_sheet_name(year)]

        self.__insert_summary_row(year_summary_sheet, self.__year_row_offset[year])

        # Name of Asset
        self._fill_cell(year_summary_sheet, self.__year_row_offset[year], 0, asset, apply_style=False)
        # Total Donations
        self._fill_cell(year_summary_sheet, self.__year_row_offset[year], 1, total_donations, apply_style=False)
        # Total Gifts
        self._fill_cell(year_summary_sheet, self.__year_row_offset[year], 2, total_gifts, apply_style=False)
        # Formula for Avg. Unit Price
        self._fill_cell(year_summary_sheet, self.__year_row_offset[year], 3, f"='{self.get_tax_sheet_name(asset, year)}'.G{row_index+10}", apply_style=False)
        # Formula for End Balance Crypto
        self._fill_cell(year_summary_sheet, self.__year_row_offset[year], 4, f"='{self.get_tax_sheet_name(asset, year)}'.I{row_index+9}", apply_style=False)
        # Formula for End Balance Yen
        self._fill_cell(year_summary_sheet, self.__year_row_offset[year], 5, f"='{self.get_tax_sheet_name(asset, year)}'.I{row_index+10}", apply_style=False)
        # Formula for Net Income Amt
        self._fill_cell(year_summary_sheet, self.__year_row_offset[year], 6, f"='{self.get_tax_sheet_name(asset, year)}'.I{row_index+18}", apply_style=False)

        self.__year_row_offset[year] += 1

        return row_index + 9
